# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import threading
from time import monotonic_ns
from typing import Any
from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.x11.bindings.core import set_context_check, X11CoreBindings
from xpra.x11.bindings.randr import RandRBindings
from xpra.x11.bindings.keyboard import X11KeyboardBindings
from xpra.x11.bindings.window import X11WindowBindings
from xpra.gtk.error import XError, xswallow, xsync, xlog, verify_sync
from xpra.gtk.util import get_default_root_window
from xpra.x11.server.server_uuid import save_uuid, get_uuid, save_mode
from xpra.x11.vfb_util import parse_resolutions
from xpra.x11.gtk.prop import prop_get, prop_set, prop_del
from xpra.x11.gtk.display_source import close_gdk_display_source
from xpra.x11.gtk.bindings import init_x11_filter, cleanup_x11_filter, cleanup_all_event_receivers
from xpra.common import MAX_WINDOW_SIZE, FULL_INFO, NotificationID
from xpra.util.objects import typedict
from xpra.util.env import envbool, first_time
from xpra.net.compression import Compressed
from xpra.net.common import PacketType
from xpra.server.gtk_server import GTKServerBase
from xpra.server import features
from xpra.x11.xkbhelper import clean_keyboard_state
from xpra.scripts.config import FALSE_OPTIONS
from xpra.log import Logger

GLib = gi_import("GLib")

set_context_check(verify_sync)
RandR = RandRBindings()
X11Keyboard = X11KeyboardBindings()
X11Core = X11CoreBindings()
X11Window = X11WindowBindings()

log = Logger("x11", "server")
keylog = Logger("x11", "server", "keyboard")
mouselog = Logger("x11", "server", "mouse")
grablog = Logger("server", "grab")
cursorlog = Logger("server", "cursor")
screenlog = Logger("server", "screen")
xinputlog = Logger("xinput")

ALWAYS_NOTIFY_MOTION = envbool("XPRA_ALWAYS_NOTIFY_MOTION", False)
FAKE_X11_INIT_ERROR = envbool("XPRA_FAKE_X11_INIT_ERROR", False)
DUMMY_WIDTH_HEIGHT_MM = envbool("XPRA_DUMMY_WIDTH_HEIGHT_MM", True)

window_type_atoms = tuple(f"_NET_WM_WINDOW_TYPE{wtype}" for wtype in (
    "",
    "_NORMAL",
    "_DESKTOP",
    "_DOCK",
    "_TOOLBAR",
    "_MENU",
    "_UTILITY",
    "_SPLASH",
    "_DIALOG",
    "_DROPDOWN_MENU",
    "_POPUP_MENU",
    "_TOOLTIP",
    "_NOTIFICATION",
    "_COMBO",
    "_DND",
    "_NORMAL"
))


class XTestPointerDevice:
    __slots__ = ()

    def __repr__(self):
        return "XTestPointerDevice"

    @staticmethod
    def move_pointer(x: int, y: int, props=None) -> None:
        mouselog("xtest_fake_motion%s", (x, y, props))
        with xsync:
            X11Keyboard.xtest_fake_motion(x, y)

    @staticmethod
    def click(button: int, pressed: bool, props: dict) -> None:
        mouselog("xtest_fake_button(%i, %s, %s)", button, pressed, props)
        with xsync:
            X11Keyboard.xtest_fake_button(button, pressed)

    @staticmethod
    def has_precise_wheel() -> bool:
        return False


# noinspection PyUnreachableCode
def get_cursor_sizes() -> tuple[int, int]:
    Gdk = gi_import("Gdk")
    display = Gdk.Display.get_default()
    if not display:
        return 0, 0
    return int(display.get_default_cursor_size()), display.get_maximal_cursor_size()


class X11ServerCore(GTKServerBase):
    """
        Base class for X11 servers,
        adds X11 specific methods to GTKServerBase.
        (see XpraServer or XpraX11ShadowServer for actual implementations)
    """

    def __init__(self) -> None:
        self.root_window = get_default_root_window()
        self.pointer_device = XTestPointerDevice()
        self.touchpad_device = None
        self.pointer_device_map: dict = {}
        self.keys_pressed: dict[int, Any] = {}
        self.x11_filter = False
        self.randr_sizes_added: list[tuple[int, int]] = []
        self.initial_resolutions: Sequence[tuple[int, int, int]] = ()
        self.randr = False
        self.randr_exact_size = False
        self.current_keyboard_group = 0
        self.key_repeat_delay = -1
        self.key_repeat_interval = -1
        super().__init__()
        log("XShape=%s", X11Window.displayHasXShape())

    def init(self, opts) -> None:
        self.do_init(opts)
        super().init(opts)

    def server_init(self) -> None:
        self.x11_init()
        if features.windows:
            from xpra.x11.window_filters import init_x11_window_filters
            init_x11_window_filters()
        super().server_init()

    def do_init(self, opts) -> None:
        onoff = sizes = opts.resize_display
        if opts.resize_display.find(":") > 0:
            #ie: "off:1080p"
            onoff, sizes = opts.resize_display.split(":", 1)
        try:
            self.initial_resolutions = parse_resolutions(sizes, opts.refresh_rate) or ()
        except ValueError:
            self.initial_resolutions = ()
        self.randr = onoff.lower() not in FALSE_OPTIONS
        self.randr_exact_size = False
        # x11 keyboard bits:
        self.current_keyboard_group = 0

    def x11_init(self) -> None:
        if FAKE_X11_INIT_ERROR:
            raise RuntimeError("fake x11 init error")
        with xlog:
            clean_keyboard_state()
        with xlog:
            if not X11Keyboard.hasXFixes() and self.cursors:
                log.error("Error: cursor forwarding support disabled")
            if not X11Keyboard.hasXTest():
                log.error("Error: keyboard and mouse disabled")
            elif not X11Keyboard.hasXkb():
                log.error("Error: limited keyboard support")
        with xsync:
            self.init_x11_atoms()
        with xlog:
            self.init_randr()
        with xlog:
            self.init_cursor()
        with xlog:
            self.x11_filter = init_x11_filter()
        assert self.x11_filter
        with xlog:
            self.save_mode()

    def save_mode(self) -> None:
        save_mode(self.get_server_mode())

    def init_randr(self) -> None:
        if self.randr and not RandR.has_randr():
            self.randr = False
        screenlog("randr=%s", self.randr)
        if not self.randr:
            return
        # check the property first,
        # because we may be inheriting this display,
        # in which case the screen sizes list may be longer than 1
        xid = self.root_window.get_xid()
        eprop = prop_get(xid, "_XPRA_RANDR_EXACT_SIZE", "u32", ignore_errors=True, raise_xerrors=False)
        screenlog("_XPRA_RANDR_EXACT_SIZE=%s", eprop)
        self.randr_exact_size = eprop == 1 or RandR.get_version() >= (1, 6)
        if not self.randr_exact_size:
            # ugly hackish way of detecting Xvfb with randr,
            # assume that it has only one resolution pre-defined:
            sizes = RandR.get_xrr_screen_sizes()
            if len(sizes) == 1:
                self.randr_exact_size = True
                prop_set(xid, "_XPRA_RANDR_EXACT_SIZE", "u32", 1)
            elif not sizes:
                # xwayland?
                self.randr = False
                self.randr_exact_size = False
        screenlog(f"randr enabled: {self.randr}, exact size={self.randr_exact_size}")
        if not self.randr:
            screenlog.warn("Warning: no X11 RandR support on %s", os.environ.get("DISPLAY"))

    def init_cursor(self) -> None:
        # cursor:
        self.default_cursor_image = None
        self.last_cursor_serial = None
        self.last_cursor_image = None
        self.send_cursor_pending = False

        def get_default_cursor():
            self.default_cursor_image = X11Keyboard.get_cursor_image()
            cursorlog("get_default_cursor=%s", self.default_cursor_image)

        with xlog:
            get_default_cursor()
            X11Keyboard.selectCursorChange(True)

    # noinspection PyMethodMayBeStatic
    def get_display_bit_depth(self) -> int:
        with xlog:
            return X11Window.get_depth(X11Window.get_root_xid())
        return 0

    # noinspection PyMethodMayBeStatic
    def init_x11_atoms(self) -> None:
        # some applications (like openoffice), do not work properly
        # if some x11 atoms aren't defined, so we define them in advance:
        X11Core.intern_atoms(window_type_atoms)

    def set_keyboard_layout_group(self, grp: int) -> None:
        kc = self.keyboard_config
        if not kc:
            keylog(f"set_keyboard_layout_group({grp}) ignored, no config")
            return
        if not kc.layout_groups:
            keylog(f"set_keyboard_layout_group({grp}) ignored, no layout groups support")
            # not supported by the client that owns the current keyboard config,
            # so make sure we stick to the default group:
            grp = 0
        if not X11Keyboard.hasXkb():
            keylog(f"set_keyboard_layout_group({grp}) ignored, no Xkb support")
            return
        if grp < 0:
            grp = 0
        if self.current_keyboard_group == grp:
            keylog(f"set_keyboard_layout_group({grp}) ignored, value unchanged")
            return
        keylog(f"set_keyboard_layout_group({grp}) config={self.keyboard_config}, {self.current_keyboard_group=}")
        try:
            with xsync:
                self.current_keyboard_group = X11Keyboard.set_layout_group(grp)
        except XError as e:
            keylog(f"set_keyboard_layout_group({grp})", exc_info=True)
            keylog.error(f"Error: failed to set keyboard layout group {grp}")
            keylog.estr(e)

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        self.add_packets("force-ungrab", "wheel-motion", main_thread=True)

    def init_virtual_devices(self, _devices) -> None:
        self.input_devices = "xtest"

    def do_cleanup(self) -> None:
        log("do_cleanup() x11_filter=%s", self.x11_filter)
        if self.x11_filter:
            self.x11_filter = False
            cleanup_x11_filter()
            # try a few times:
            # errors happen because windows are being destroyed
            # (even more so when we call `cleanup`)
            # and we don't really care too much about this
            for log_fn in (log.debug, log.debug, log.debug, log.debug, log.warn):
                try:
                    with xsync:
                        cleanup_all_event_receivers()
                        # all went well, we're done
                        log("all event receivers have been removed")
                        break
                except Exception as e:
                    log("do_cleanup() cleanup_all_event_receivers()", exc_info=True)
                    log_fn("failed to remove event receivers: %s", e)
        with xlog:
            clean_keyboard_state()
        # prop_del does its own xsync:
        self.clean_x11_properties()
        super().do_cleanup()
        log("close_gdk_display_source()")
        close_gdk_display_source()

    def clean_x11_properties(self) -> None:
        self.do_clean_x11_properties("XPRA_SERVER_MODE", "_XPRA_RANDR_EXACT_SIZE")

    # noinspection PyMethodMayBeStatic
    def do_clean_x11_properties(self, *properties) -> None:
        root_xid = X11Window.get_root_xid()
        for prop in properties:
            try:
                prop_del(root_xid, prop)
            except Exception as e:
                log.warn(f"Warning: failed to delete property {prop!r} on root window {root_xid}: {e}")

    # noinspection PyMethodMayBeStatic
    def get_uuid(self) -> str:
        return get_uuid()

    def save_uuid(self) -> None:
        save_uuid(str(self.uuid))

    def set_keyboard_repeat(self, key_repeat) -> None:
        with xlog:
            if key_repeat:
                self.key_repeat_delay, self.key_repeat_interval = key_repeat
                if self.key_repeat_delay > 0 and self.key_repeat_interval > 0:
                    X11Keyboard.set_key_repeat_rate(self.key_repeat_delay, self.key_repeat_interval)
                    keylog.info("setting key repeat rate from client: %sms delay / %sms interval",
                                self.key_repeat_delay, self.key_repeat_interval)
            else:
                # dont do any jitter compensation:
                self.key_repeat_delay = -1
                self.key_repeat_interval = -1
                # but do set a default repeat rate:
                X11Keyboard.set_key_repeat_rate(500, 30)
                keylog("keyboard repeat disabled")

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = super().make_hello(source)
        capabilities["server_type"] = "Python/gtk/x11"
        if "features" in source.wants:
            capabilities |= {
                "resize_screen": self.randr,
                "resize_exact": self.randr_exact_size,
                "force_ungrab": True,
                "keyboard.fast-switching": True,
                "wheel.precise": self.pointer_device.has_precise_wheel(),
                "pointer.optional": True,
                "touchpad-device": bool(self.touchpad_device),
            }
            if self.randr:
                sizes = self.get_all_screen_sizes()
                if len(sizes) > 1:
                    capabilities["screen-sizes"] = sizes
            if self.default_cursor_image and "default_cursor" in source.wants:
                ce = getattr(source, "cursor_encodings", ())
                if "default" not in ce:
                    # we have to send it this way
                    # instead of using send_initial_cursors()
                    capabilities["cursor.default"] = self.default_cursor_image
        return capabilities

    def send_initial_cursors(self, ss, sharing: bool = False) -> None:
        dci = self.default_cursor_image
        encodings = getattr(ss, "cursor_encodings", ())
        enabled = getattr(self, "cursors", False)
        if not (enabled and ("default" in encodings) and dci):
            return
        cursorlog(f"default_cursor_image={dci}, {enabled=}, {encodings=}", )
        with cursorlog.trap_error("Error sending default cursor"):
            ss.do_send_cursor(0, dci, get_cursor_sizes(), encoding_prefix="default:")

    def do_get_info(self, proto, server_sources) -> dict[str, Any]:
        start = monotonic_ns()
        info = super().do_get_info(proto, server_sources)
        sinfo = info.setdefault("server", {})
        sinfo["type"] = "Python/gtk/x11"
        if FULL_INFO > 1:
            try:
                from xpra.codecs.drm.drm import query  # pylint: disable=import-outside-toplevel
            except ImportError as e:
                log(f"no drm query: {e}")
            else:
                sinfo["drm"] = query()
        log("X11ServerCore.do_get_info took %ims", (monotonic_ns() - start) // 1_000_000)
        return info

    def get_ui_info(self, proto, wids=None, *args) -> dict[str, Any]:
        log("do_get_info thread=%s", threading.current_thread())
        info = super().get_ui_info(proto, wids, *args)
        # this is added here because the server keyboard config doesn't know about "keys_pressed"..
        if not self.readonly:
            with xlog:
                info.setdefault("keyboard", {}).update(
                    {
                        "state":
                            {
                                "keys_pressed": tuple(self.keys_pressed.keys()),
                                "keycodes-down": X11Keyboard.get_keycodes_down(),
                            },
                        "fast-switching": True,
                        "layout-group": X11Keyboard.get_layout_group(),
                    }
                )
        sinfo = info.setdefault("server", {})
        try:
            from xpra.x11.gtk.composite import CompositeHelper
            sinfo["XShm"] = CompositeHelper.XShmEnabled
        except ImportError:
            pass
        # cursor:
        if self.last_cursor_image:
            info.setdefault("cursor", {}).update({"current": self.get_cursor_info()})
        with xswallow:
            sinfo |= {
                "Xkb": X11Keyboard.hasXkb(),
                "XTest": X11Keyboard.hasXTest(),
            }
        # randr:
        if self.randr:
            with xlog:
                sizes = self.get_all_screen_sizes()
                if sizes:
                    sinfo["randr"] = {
                        "": True,
                        "options": tuple(reversed(sorted(sizes))),
                        "exact": self.randr_exact_size,
                    }
        return info

    def get_cursor_info(self) -> dict[str, Any]:
        # (NOT from UI thread)
        # copy to prevent race:
        cd = self.last_cursor_image
        if cd is None:
            return {}
        dci = self.default_cursor_image
        cinfo = {
            "is-default": bool(dci) and len(dci) >= 8 and len(cd) >= 8 and cd[7] == dci[7],
        }
        # all but pixels:
        for i, x in enumerate(("x", "y", "width", "height", "xhot", "yhot", "serial", None, "name")):
            if x:
                v = cd[i] or ""
                cinfo[x] = v
        return cinfo

    def get_window_info(self, window) -> dict[str, Any]:
        info = super().get_window_info(window)
        info["XShm"] = window.uses_xshm()
        info["geometry"] = window.get_geometry()
        return info

    def get_window_id(self, gdkwindow) -> int:
        if not gdkwindow:
            return 0
        xid = gdkwindow.get_xid()
        if not xid:
            return 0
        for wid, window in self._id_to_window.items():
            if window.get("xid", 0) == xid:
                return wid
        return 0

    # noinspection PyMethodMayBeStatic
    def get_keyboard_config(self, props=None):
        p = props or typedict()
        from xpra.x11.server.keyboard_config import KeyboardConfig
        keyboard_config = KeyboardConfig()
        keyboard_config.enabled = p.boolget("keyboard", True)
        keyboard_config.parse_options(p)
        keyboard_config.parse_layout(p)
        keylog("get_keyboard_config(..)=%s", keyboard_config)
        return keyboard_config

    def set_keymap(self, server_source, force=False) -> None:
        if self.readonly:
            return

        def reenable_keymap_changes(*args) -> bool:
            keylog("reenable_keymap_changes(%s)", args)
            self.keymap_changing_timer = 0
            self._keys_changed()
            return False

        # prevent _keys_changed() from firing:
        # (using a flag instead of keymap.disconnect(handler) as this did not seem to work!)
        if not self.keymap_changing_timer:
            # use idle_add to give all the pending
            # events a chance to run first (and get ignored)
            self.keymap_changing_timer = GLib.timeout_add(100, reenable_keymap_changes)
        # if sharing, don't set the keymap, translate the existing one:
        other_ui_clients = [s.uuid for s in self._server_sources.values() if s != server_source and s.ui_client]
        translate_only = len(other_ui_clients) > 0
        keylog("set_keymap(%s, %s) translate_only=%s", server_source, force, translate_only)
        with xsync:
            # pylint: disable=access-member-before-definition
            server_source.set_keymap(self.keyboard_config, self.keys_pressed, force, translate_only)
            self.keyboard_config = server_source.keyboard_config

    def clear_keys_pressed(self) -> None:
        if self.readonly:
            return
        keylog("clear_keys_pressed()")
        # make sure the timer doesn't fire and interfere:
        self.cancel_key_repeat_timer()
        # clear all the keys we know about:
        if self.keys_pressed:
            keylog("clearing keys pressed: %s", self.keys_pressed)
            with xsync:
                for keycode in self.keys_pressed:
                    self.fake_key(keycode, False)
            self.keys_pressed = {}
        # this will take care of any remaining ones we are not aware of:
        # (there should not be any - but we want to be certain)
        clean_keyboard_state()

    # noinspection PyMethodMayBeStatic
    def get_cursor_image(self):
        # must be called from the UI thread!
        with xlog:
            return X11Keyboard.get_cursor_image()

    def get_cursor_data(self, skip_default=True) -> tuple[Any, Any]:
        # must be called from the UI thread!
        cursor_image = self.get_cursor_image()
        if cursor_image is None:
            cursorlog("get_cursor_data() failed to get cursor image")
            return None, []
        self.last_cursor_image = list(cursor_image)
        pixels = self.last_cursor_image[7]
        cursorlog("get_cursor_image() cursor=%s", cursor_image[:7] + ["%s bytes" % len(pixels)] + cursor_image[8:])
        is_default = self.default_cursor_image is not None and str(pixels) == str(self.default_cursor_image[7])
        if skip_default and is_default:
            cursorlog("get_cursor_data(): default cursor - clearing it")
            cursor_image = None
        cursor_sizes = get_cursor_sizes()
        return cursor_image, cursor_sizes

    def get_all_screen_sizes(self) -> Sequence[tuple[int, int]]:
        # workaround for #2910: the resolutions we add are not seen by XRRSizes!
        # so we keep track of the ones we have added ourselves:
        sizes = list(RandR.get_xrr_screen_sizes())
        for w, h in self.randr_sizes_added:
            if (w, h) not in sizes:
                sizes.append((w, h))
        return tuple(sizes)

    def get_max_screen_size(self) -> tuple[int, int]:
        max_w, max_h = self.root_window.get_geometry()[2:4]
        if self.randr:
            sizes = self.get_all_screen_sizes()
            if len(sizes) >= 1:
                for w, h in sizes:
                    max_w = max(max_w, w)
                    max_h = max(max_h, h)
            if max_w > MAX_WINDOW_SIZE or max_h > MAX_WINDOW_SIZE:
                screenlog.warn("Warning: maximum screen size is very large: %sx%s", max_w, max_h)
                screenlog.warn(" you may encounter window sizing problems")
            screenlog("get_max_screen_size()=%s", (max_w, max_h))
        return max_w, max_h

    def configure_best_screen_size(self):
        # return ServerBase.set_best_screen_size(self)
        """ sets the screen size to use the largest width and height used by any of the clients """
        root_w, root_h = self.root_window.get_geometry()[2:4]
        if not self.randr:
            return root_w, root_h
        sss = tuple(x for x in self._server_sources.values() if x.ui_client)
        max_w, max_h = 0, 0
        min_w, min_h = 16384, 16384
        client_sizes = {}
        for ss in sss:
            client_size = ss.desktop_size
            if client_size:
                w, h = client_size
                size = "%ix%i" % (w, h)
                max_w = max(max_w, w)
                max_h = max(max_h, h)
                if w > 0:
                    min_w = min(min_w, w)
                if h > 0:
                    min_h = min(min_h, h)
                client_sizes[ss.uuid] = size
        if len(client_sizes) > 1:
            screenlog.info("screen used by %i clients:", len(client_sizes))
            for uuid, size in client_sizes.items():
                screenlog.info("* %s: %s", uuid, size)
        screenlog("current server resolution is %ix%i", root_w, root_h)
        screenlog("maximum client resolution is %ix%i", max_w, max_h)
        screenlog("minimum client resolution is %ix%i", min_w, min_h)
        w, h = max_w, max_h
        screenlog("using %ix%i", w, h)
        if w <= 0 or h <= 0:
            # invalid - use fallback
            return root_w, root_h
        return self.set_screen_size(w, h)

    def get_best_screen_size(self, desired_w: int, desired_h: int):
        r = self.do_get_best_screen_size(desired_w, desired_h)
        screenlog("get_best_screen_size%s=%s", (desired_w, desired_h), r)
        return r

    def do_get_best_screen_size(self, desired_w: int, desired_h: int):
        if not self.randr:
            return desired_w, desired_h
        screen_sizes = self.get_all_screen_sizes()
        if (desired_w, desired_h) in screen_sizes:
            return desired_w, desired_h
        if self.randr_exact_size:
            try:
                with xsync:
                    if RandR.add_screen_size(desired_w, desired_h):
                        # we have to wait a little bit
                        # to make sure that everything sees the new resolution
                        # (ideally this method would be split in two and this would be a callback)
                        self.randr_sizes_added.append((desired_w, desired_h))
                        import time
                        time.sleep(0.5)
                        return desired_w, desired_h
            except XError as e:
                screenlog("add_screen_size(%s, %s)", desired_w, desired_h, exc_info=True)
                screenlog.warn("Warning: failed to add resolution %ix%i:", desired_w, desired_h)
                screenlog.warn(" %s", e)
            # re-query:
            screen_sizes = self.get_all_screen_sizes()
        # try to find the best screen size to resize to:
        closest = {}
        for w, h in screen_sizes:
            distance = abs(desired_w * desired_h - w * h)
            closest[distance] = (w, h)
        if not closest:
            screenlog.warn("Warning: no matching resolution found for %sx%s", desired_w, desired_h)
            root_w, root_h = self.root_window.get_size()
            return root_w, root_h
        min_dist = sorted(closest.keys())[0]
        new_size = closest[min_dist]
        screenlog("best %s resolution for client(%sx%s) is: %s", desired_w, desired_h, new_size)
        w, h = new_size
        return w, h

    def set_screen_size(self, desired_w: int, desired_h: int):
        screenlog("set_screen_size%s", (desired_w, desired_h))
        root_w, root_h = self.root_window.get_geometry()[2:4]
        if not self.randr:
            return root_w, root_h
        if desired_w == root_w and desired_h == root_h:
            return root_w, root_h  # unlikely: perfect match already!
        # clients may supply "xdpi" and "ydpi" (v0.15 onwards), or just "dpi", or nothing...
        xdpi = self.xdpi or self.dpi
        ydpi = self.ydpi or self.dpi
        screenlog("set_screen_size(%s, %s) xdpi=%s, ydpi=%s",
                  desired_w, desired_h, xdpi, ydpi)
        wmm, hmm = 0, 0
        if xdpi <= 0 or ydpi <= 0:
            # use some sane defaults: either the command line option, or fallback to 96
            # (96 is better than nothing, because we do want to set the dpi
            # to avoid Xdummy setting a crazy dpi from the virtual screen dimensions)
            xdpi = self.default_dpi or 96
            ydpi = self.default_dpi or 96
            # find the "physical" screen dimensions, so we can calculate the required dpi
            # (and do this before changing the resolution)
            client_w, client_h = 0, 0
            sss = self._server_sources.values()
            for ss in sss:
                screen_sizes = getattr(ss, "screen_sizes", ())
                for s in screen_sizes:
                    if len(s) >= 10:
                        # (display_name, width, height, width_mm, height_mm, monitors,
                        # work_x, work_y, work_width, work_height)
                        client_w = max(client_w, s[1])
                        client_h = max(client_h, s[2])
                        wmm = max(wmm, s[3])
                        hmm = max(hmm, s[4])
            if wmm > 0 and hmm > 0 and client_w > 0 and client_h > 0:
                # calculate "real" dpi:
                xdpi = round(client_w * 25.4 / wmm)
                ydpi = round(client_h * 25.4 / hmm)
                screenlog("calculated DPI: %s x %s (from w: %s / %s, h: %s / %s)",
                          xdpi, ydpi, client_w, wmm, client_h, hmm)
        if wmm == 0 or hmm == 0:
            wmm = round(desired_w * 25.4 / xdpi)
            hmm = round(desired_h * 25.4 / ydpi)
        if DUMMY_WIDTH_HEIGHT_MM:
            # FIXME: we assume there is only one output:
            output = 0
            with xsync:
                RandR.set_output_int_property(output, "WIDTH_MM", wmm)
                RandR.set_output_int_property(output, "HEIGHT_MM", hmm)
        screenlog("set_dpi(%i, %i)", xdpi, ydpi)
        self.set_dpi(xdpi, ydpi)

        # try to find the best screen size to resize to:
        w, h = self.get_best_screen_size(desired_w, desired_h)

        if w == root_w and h == root_h:
            screenlog.info("best resolution matching %sx%s is unchanged: %sx%s", desired_w, desired_h, w, h)
            return root_w, root_h
        with screenlog.trap_error("Error: failed to set new resolution"):
            with xsync:
                RandR.get_screen_size()
            # Xdummy with randr 1.2:
            screenlog("using XRRSetScreenConfigAndRate with %ix%i", w, h)
            with xsync:
                RandR.set_screen_size(w, h)
            if self.randr_exact_size:
                # Xvfb with randr > 1.2: the resolution has been added
                # we can use XRRSetScreenSize:
                try:
                    with xsync:
                        RandR.xrr_set_screen_size(w, h, wmm, hmm)
                except XError:
                    screenlog("XRRSetScreenSize failed", exc_info=True)
            screenlog("calling RandR.get_screen_size()")
            with xsync:
                root_w, root_h = RandR.get_screen_size()
            screenlog("RandR.get_screen_size()=%s,%s", root_w, root_h)
            screenlog("RandR.get_vrefresh()=%s", RandR.get_vrefresh())
            if root_w != w or root_h != h:
                screenlog.warn("Warning: tried to set resolution to %ix%i", w, h)
                screenlog.warn(" and ended up with %ix%i", root_w, root_h)
            else:
                msg = f"server virtual display now set to {root_w}x{root_h}"
                if desired_w != root_w or desired_h != root_h:
                    msg += f" (best match for {desired_w}x{desired_h})"
                screenlog.info(msg)

            # show dpi via idle_add so server has time to change the screen size (mm)
            GLib.idle_add(self.show_dpi, xdpi, ydpi)
        return root_w, root_h

    def show_dpi(self, xdpi: int, ydpi: int):
        root_w, root_h = self.root_window.get_geometry()[2:4]
        wmm, hmm = RandR.get_screen_size_mm()  # ie: (1280, 1024)
        screenlog("RandR.get_screen_size_mm=%s,%s", wmm, hmm)
        actual_xdpi = round(root_w * 25.4 / wmm)
        actual_ydpi = round(root_h * 25.4 / hmm)
        if abs(actual_xdpi - xdpi) <= 1 and abs(actual_ydpi - ydpi) <= 1:
            screenlog.info("DPI set to %s x %s", actual_xdpi, actual_ydpi)
            screenlog("wanted: %s x %s", xdpi, ydpi)
        else:
            # should this be a warning:
            log_fn = screenlog.info
            maxdelta = max(abs(actual_xdpi - xdpi), abs(actual_ydpi - ydpi))
            if maxdelta >= 10:
                log_fn = screenlog.warn
            messages = [
                f"DPI set to {actual_xdpi} x {actual_ydpi} (wanted {xdpi} x {ydpi})",
            ]
            if maxdelta >= 10:
                messages.append("you may experience scaling problems, such as huge or small fonts, etc")
                messages.append("to fix this issue, try the dpi switch, or use a patched Xorg dummy driver")
                self.notify_dpi_warning("\n".join(messages))
            for i, message in enumerate(messages):
                log_fn("%s%s", ["", " "][i > 0], message)

    def mirror_client_monitor_layout(self) -> dict[int, Any]:
        with xsync:
            assert RandR.is_dummy16(), "cannot match monitor layout without RandR 1.6"
        # if we have a single UI client,
        # see if we can emulate its monitor geometry exactly
        sss = tuple(x for x in self._server_sources.values() if x.ui_client)
        screenlog("%i sources=%s", len(sss), sss)
        if len(sss) != 1:
            return {}
        ss = sss[0]
        mdef = ss.get_monitor_definitions()
        if not mdef:
            return {}
        screenlog(f"monitor definition from client {ss}: {mdef}")
        from xpra.common import adjust_monitor_refresh_rate
        mdef = adjust_monitor_refresh_rate(self.refresh_rate, mdef)
        screenlog("refresh-rate adjusted using %s: %s", self.refresh_rate, mdef)
        with xlog:
            RandR.set_crtc_config(mdef)
        return mdef

    def notify_dpi_warning(self, body: str) -> None:
        sources = tuple(self._server_sources.values())
        if len(sources) == 1:
            ss = sources[0]
            if first_time("DPI-warning-%s" % ss.uuid):
                sources[0].may_notify(NotificationID.DPI, "DPI Issue", body, icon_name="font")

    def set_dpi(self, xdpi: int, ydpi: int) -> None:
        """ overridden in the seamless server """

    def _process_server_settings(self, _proto, packet: PacketType) -> None:
        settings = packet[1]
        log("process_server_settings: %s", settings)
        self.update_server_settings(settings)

    def update_server_settings(self, _settings, _reset=False) -> None:
        # implemented in the X11 xpra server only for now
        # (does not make sense to update a shadow server)
        log("ignoring server settings update in %s", self)

    def _process_force_ungrab(self, proto, _packet: PacketType) -> None:
        # ignore the window id: wid = packet[1]
        grablog("force ungrab from %s", proto)
        self.X11_ungrab()

    # noinspection PyMethodMayBeStatic
    def X11_ungrab(self) -> None:
        grablog("X11_ungrab")
        with xsync:
            X11Core.UngrabKeyboard()
            X11Core.UngrabPointer()

    # noinspection PyMethodMayBeStatic
    def fake_key(self, keycode, press) -> None:
        keylog("fake_key(%s, %s)", keycode, press)
        mink, maxk = X11Keyboard.get_minmax_keycodes()
        if keycode < mink or keycode > maxk:
            return
        with xsync:
            X11Keyboard.xtest_fake_key(keycode, press)

    def do_x11_cursor_event(self, event) -> None:
        cursors = getattr(self, "cursors", False)
        if not cursors:
            return
        if self.last_cursor_serial == event.cursor_serial:
            cursorlog("ignoring cursor event %s with the same serial number %s", event, self.last_cursor_serial)
            return
        cursorlog("cursor_event: %s", event)
        self.last_cursor_serial = event.cursor_serial
        for ss in self.window_sources():
            ss.send_cursor()

    def _motion_signaled(self, model, event) -> None:
        mouselog("motion_signaled(%s, %s) last mouse user=%s", model, event, self.last_mouse_user)
        # find the window model for this gdk window:
        wid = self._window_to_id.get(model)
        if not wid:
            return
        for ss in self._server_sources.values():
            if ALWAYS_NOTIFY_MOTION or self.last_mouse_user is None or self.last_mouse_user != ss.uuid:
                if hasattr(ss, "update_mouse"):
                    ss.update_mouse(wid, event.x_root, event.y_root, event.x, event.y)

    def do_x11_xkb_event(self, event) -> None:
        # X11: XKBNotify
        log("server do_x11_xkb_event(%r)" % event)
        if event.subtype != "bell":
            log.error(f"Error: unknown event subtype: {event.subtype!r}")
            log.error(f" {event=}")
            return
        # bell events on our windows will come through the bell signal,
        # this method is a catch-all for events on windows we don't manage,
        # so we use wid=0 for that:
        wid = 0
        for ss in self.window_sources():
            ss.bell(wid, event.device, event.percent,
                    event.pitch, event.duration, event.bell_class, event.bell_id, event.bell_name)

    def _bell_signaled(self, wm, event) -> None:
        log("bell signaled on window %#x", event.window)
        if not self.bell:
            return
        wid = 0
        rxid = X11Window.get_root_xid()
        if event.window != rxid and event.window_model is not None:
            wid = self._window_to_id.get(event.window_model, 0)
        log("_bell_signaled(%s,%r) wid=%s", wm, event, wid)
        for ss in self.window_sources():
            ss.bell(wid, event.device, event.percent,
                    event.pitch, event.duration, event.bell_class, event.bell_id, event.bell_name)

    def setup_input_devices(self) -> None:
        xinputlog("setup_input_devices() input_devices feature=%s", features.input_devices)
        if not features.input_devices:
            return
        xinputlog("setup_input_devices() format=%s, input_devices=%s", self.input_devices_format, self.input_devices)
        xinputlog("setup_input_devices() input_devices_data=%s", self.input_devices_data)
        # xinputlog("setup_input_devices() input_devices_data=%s", self.input_devices_data)
        xinputlog("setup_input_devices() pointer device=%s", self.pointer_device)
        xinputlog("setup_input_devices() touchpad device=%s", self.touchpad_device)
        self.pointer_device_map = {}
        if not self.touchpad_device:
            # no need to assign anything, we only have one device anyway
            return
        # if we find any absolute pointer devices,
        # map them to the "touchpad_device"
        XIModeAbsolute = 1
        for deviceid, device_data in self.input_devices_data.items():
            name = device_data.get("name")
            # xinputlog("[%i]=%s", deviceid, device_data)
            xinputlog("[%i]=%s", deviceid, name)
            if device_data.get("use") != "slave pointer":
                continue
            classes = device_data.get("classes")
            if not classes:
                continue
            # look for absolute pointer devices:
            touchpad_axes = []
            for i, defs in classes.items():
                xinputlog(" [%i]=%s", i, defs)
                mode = defs.get("mode")
                label = defs.get("label")
                if not mode or mode != XIModeAbsolute:
                    continue
                if defs.get("min", -1) == 0 and defs.get("max", -1) == (2 ** 24 - 1):
                    touchpad_axes.append((i, label))
            if len(touchpad_axes) == 2:
                xinputlog.info("found touchpad device: %s", name)
                xinputlog("axes: %s", touchpad_axes)
                self.pointer_device_map[deviceid] = self.touchpad_device

    def _process_wheel_motion(self, proto, packet: PacketType) -> None:
        assert self.pointer_device.has_precise_wheel()
        ss = self.get_server_source(proto)
        if not ss:
            return
        wid, button, distance, pointer, modifiers, _buttons = packet[1:7]
        device_id = -1
        props = {}
        self.record_wheel_event(wid, button)
        with xsync:
            if self.do_process_mouse_common(proto, device_id, wid, pointer, props):
                self.last_mouse_user = ss.uuid
                self._update_modifiers(proto, wid, modifiers)
                self.pointer_device.wheel_motion(button, distance / 1000.0)  # pylint: disable=no-member

    def get_pointer_device(self, deviceid: int):
        # mouselog("get_pointer_device(%i) input_devices_data=%s", deviceid, self.input_devices_data)
        if self.input_devices_data:
            device_data = self.input_devices_data.get(deviceid)
            if device_data:
                mouselog("get_pointer_device(%i) device=%s", deviceid, device_data.get("name"))
        device = self.pointer_device_map.get(deviceid) or self.pointer_device
        return device

    def _get_pointer_abs_coordinates(self, wid: int, pos) -> tuple[int, int]:
        # simple absolute coordinates
        x, y = pos[:2]
        from xpra.server.mixins.window import WindowServer
        if len(pos) >= 4 and isinstance(self, WindowServer):
            # relative coordinates
            model = self._id_to_window.get(wid)
            if model:
                rx, ry = pos[2:4]
                geom = model.get_geometry()
                x = geom[0] + rx
                y = geom[1] + ry
                mouselog("_get_pointer_abs_coordinates(%i, %s)=%s window geometry=%s", wid, pos, (x, y), geom)
        return x, y

    def _move_pointer(self, device_id: int, wid: int, pos, props=None) -> None:
        # (this is called within a `xswallow` context)
        x, y = self._get_pointer_abs_coordinates(wid, pos)
        self.device_move_pointer(device_id, wid, (x, y), props)

    def device_move_pointer(self, device_id: int, wid: int, pos, props):
        device = self.get_pointer_device(device_id)
        x, y = pos
        mouselog("move_pointer(%s, %s, %s) device=%s, position=%s",
                 wid, pos, device_id, device, (x, y))
        try:
            device.move_pointer(x, y, props)
        except Exception as e:
            mouselog.error("Error: failed to move the pointer to %sx%s using %s", x, y, device)
            mouselog.estr(e)

    def do_process_mouse_common(self, proto, device_id: int, wid: int, pointer, props) -> bool:
        mouselog("do_process_mouse_common%s", (proto, device_id, wid, pointer, props))
        if self.readonly:
            return False
        with xsync:
            pos = X11Keyboard.query_pointer()
        if (pointer and pos != pointer[:2]) or self.input_devices == "xi":
            with xswallow:
                self._move_pointer(device_id, wid, pointer, props)
        return True

    def _update_modifiers(self, proto, wid: int, modifiers) -> None:
        if self.readonly:
            return
        ss = self.get_server_source(proto)
        if ss:
            if self.ui_driver and self.ui_driver != ss.uuid:
                return
            if hasattr(ss, "keyboard_config"):
                ss.make_keymask_match(modifiers)
            if wid == self.get_focus():
                ss.user_event()

    def do_process_button_action(self, proto, device_id: int, wid: int, button: int, pressed: bool,
                                 pointer, props: dict) -> None:
        if "modifiers" in props:
            self._update_modifiers(proto, wid, props.get("modifiers"))
        props = {}
        if self.process_mouse_common(proto, device_id, wid, pointer, props):
            self.button_action(device_id, wid, pointer, button, pressed, props)

    def button_action(self, device_id: int, wid: int, pointer, button: int, pressed: bool, props: dict) -> None:
        device = self.get_pointer_device(device_id)
        assert device, "pointer device %s not found" % device_id
        if button in (4, 5) and wid:
            self.record_wheel_event(wid, button)
        try:
            mouselog("%s%s", device.click, (button, pressed, props))
            with xsync:
                device.click(button, pressed, props)
        except XError:
            mouselog("button_action%s", (device_id, wid, pointer, button, pressed, props), exc_info=True)
            mouselog.error("Error: failed (un)press mouse button %s", button)

    def record_wheel_event(self, wid: int, button: int) -> None:
        mouselog("recording scroll event for button %i", button)
        for ss in self.window_sources():
            ss.record_scroll_event(wid)

    @staticmethod
    def make_screenshot_packet_from_regions(regions) -> tuple[str, int, int, str, int, Any]:
        # regions = array of (wid, x, y, PIL.Image)
        if not regions:
            log("screenshot: no regions found, returning empty 0x0 image!")
            return "screenshot", 0, 0, "png", 0, b""
        # in theory, we could run the rest in a non-UI thread since we're done with GTK..
        minx: int = min(x for (_, x, _, _) in regions)
        miny: int = min(y for (_, _, y, _) in regions)
        maxx: int = max((x + img.get_width()) for (_, x, _, img) in regions)
        maxy: int = max((y + img.get_height()) for (_, _, y, img) in regions)
        width = maxx - minx
        height = maxy - miny
        log("screenshot: %sx%s, min x=%s y=%s", width, height, minx, miny)
        from PIL import Image  # pylint: disable=import-outside-toplevel
        screenshot = Image.new("RGBA", (width, height))
        for wid, x, y, img in reversed(regions):
            pixel_format = img.get_pixel_format()
            target_format = {
                "XRGB": "RGB",
                "BGRX": "RGB",
                "BGRA": "RGBA",
            }.get(pixel_format, pixel_format)
            pixels = img.get_pixels()
            w = img.get_width()
            h = img.get_height()
            # PIL cannot use the memoryview directly:
            if isinstance(pixels, memoryview):
                pixels = pixels.tobytes()
            try:
                window_image = Image.frombuffer(target_format, (w, h), pixels, "raw", pixel_format, img.get_rowstride())
            except (ValueError, TypeError):
                log.error("Error parsing window pixels in %s format for window %i", pixel_format, wid, exc_info=True)
                continue
            tx = x - minx
            ty = y - miny
            screenshot.paste(window_image, (tx, ty))
        from io import BytesIO
        buf = BytesIO()
        screenshot.save(buf, "png")
        data = buf.getvalue()
        buf.close()
        packet = ("screenshot", width, height, "png", width * 4, Compressed("png", data))
        log("screenshot: %sx%s %s", packet[1], packet[2], packet[-1])
        return packet



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.os_util import OSX, WIN32, gi_import
from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.log import Logger

Gtk = gi_import("Gtk")

log = Logger("util")


def has_client() -> bool:
    try:
        from xpra import client
        return bool(client)
    except ImportError:
        return False


def has_mdns() -> bool:
    try:
        from xpra.net.mdns import util
        return bool(util)
    except ImportError:
        return False


def has_server() -> bool:
    try:
        from xpra.server import util
        return bool(util)
    except ImportError:
        return False


def has_shadow() -> bool:
    try:
        from xpra.server import shadow
        return bool(shadow)
    except ImportError:
        return False


def has_configure() -> bool:
    try:
        from xpra.gtk import configure
        return bool(configure)
    except ImportError:
        return False


class GUI(BaseGUIWindow):

    def __init__(self, argv=()):
        self.argv = argv
        self.widgets = []
        super().__init__(header_bar=(True, True, True))

    def populate(self) -> None:
        def browse_tooltip() -> str:
            if not has_client():
                return "the client is not installed"
            if not has_mdns():
                return "the mdns module is not installed"
            return "Browse and connect to local and mDNS sessions"
        self.ib("Browse", "browse.png", browse_tooltip(), self.browse, sensitive=has_client() and has_mdns())

        def connect_tooltip() -> str:
            if not has_client():
                return "the client is not installed"
            return "Connect to an existing session\nover the network"
        self.ib("Connect", "connect.png", connect_tooltip(), self.show_launcher, sensitive=has_client())

        def shadow_tooltip() -> str:
            if not has_shadow():
                return "the shadow server feature is not installed"
            return "\n".join((
                "Start a shadow server,",
                "making this desktop accessible to others",
                "(authentication required)",
            ))
        self.ib("Shadow", "server-connected.png", shadow_tooltip(), self.shadow, sensitive=has_shadow())

        self.ib("Configure", "ticked.png", "", self.configure, sensitive=has_configure())

        def start_tooltip() -> str:
            if not has_client():
                return "the client is not installed"
            return "Start a new %sxpra session" % ("remote " if not has_server() else "")
        self.ib("Start", "windows.png", start_tooltip(), self.start, sensitive=has_client())

        grid = Gtk.Grid()
        grid.set_row_homogeneous(True)
        grid.set_column_homogeneous(True)
        for i, widget in enumerate(self.widgets):
            grid.attach(widget, i % 2, i // 2, 1, 1)
        self.vbox.add(grid)

    def add_widget(self, widget) -> None:
        self.widgets.append(widget)

    def get_xpra_command(self, *args) -> list[str]:
        argv = list(self.argv[1:])
        if argv.index("gui") >= 0:
            argv.pop(argv.index("gui"))
        return super().get_xpra_command(*args) + argv

    def configure(self, button) -> None:
        self.button_command(button, "configure")

    def shadow(self, button) -> None:
        cmd_args = ["shadow", "--bind-tcp=0.0.0.0:14500,auth=sys,ssl-cert=auto"] if (WIN32 or OSX) else ["shadow"]
        self.button_command(button, *cmd_args)

    def browse(self, btn) -> None:
        self.button_command(btn, "sessions")

    def show_launcher(self, btn) -> None:
        self.button_command(btn, "launcher")

    def start(self, btn) -> None:
        self.button_command(btn, "start-gui")

    def open_file(self, filename: str) -> None:
        log("open_file(%s)", filename)
        self.exec_subcommand("launcher", filename)

    def open_url(self, url: str) -> None:
        log("open_url(%s)", url)
        self.exec_subcommand("attach", url)


def main(argv):  # pragma: no cover
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.log import enable_color
    from xpra.platform.gui import init, ready
    from xpra.gtk.signals import register_os_signals
    with program_context("xpra-gui", "Xpra GUI"):
        enable_color()
        init()
        gui = GUI(argv=argv)
        register_os_signals(gui.app_signal)
        ready()
        if OSX:
            from xpra.platform.darwin.gui import wait_for_open_handlers
            wait_for_open_handlers(gui.show, gui.open_file, gui.open_url)
        else:
            gui.show()
        Gtk.main()
        log("do_main() gui.exit_code=%i", gui.exit_code)
        return 0


if __name__ == "__main__":  # pragma: no cover
    r = main(sys.argv)
    sys.exit(r)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


def get_client_base_classes() -> tuple[type, ...]:
    from xpra.client.base.client import XpraClientBase
    from xpra.client.base import features

    CLIENT_BASES: list[type] = [XpraClientBase]
    if features.display:
        from xpra.client.mixins.display import DisplayClient
        CLIENT_BASES.append(DisplayClient)
    if features.windows:
        from xpra.client.mixins.windows import WindowClient
        CLIENT_BASES.append(WindowClient)
        if features.cursors:
            from xpra.client.mixins.cursors import CursorClient
            CLIENT_BASES.append(CursorClient)
    if features.webcam:
        from xpra.client.mixins.webcam import WebcamForwarder
        CLIENT_BASES.append(WebcamForwarder)
    if features.audio:
        from xpra.client.mixins.audio import AudioClient
        CLIENT_BASES.append(AudioClient)
    if features.clipboard:
        from xpra.client.mixins.clipboard import ClipboardClient
        CLIENT_BASES.append(ClipboardClient)
    if features.notifications:
        from xpra.client.mixins.notification import NotificationClient
        CLIENT_BASES.append(NotificationClient)
    if features.mmap:
        from xpra.client.mixins.mmap import MmapClient
        CLIENT_BASES.append(MmapClient)
    if features.logging:
        from xpra.client.mixins.logging import RemoteLogging
        CLIENT_BASES.append(RemoteLogging)
    if features.network_state:
        from xpra.client.mixins.network_state import NetworkState
        CLIENT_BASES.append(NetworkState)
    if features.network_listener:
        from xpra.client.mixins.network_listener import NetworkListener
        CLIENT_BASES.append(NetworkListener)
    if features.encoding:
        from xpra.client.mixins.encodings import Encodings
        CLIENT_BASES.append(Encodings)
    if features.tray:
        from xpra.client.mixins.tray import TrayClient
        CLIENT_BASES.append(TrayClient)

    return tuple(CLIENT_BASES)



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from typing import Any

from SystemConfiguration import (
    SCNetworkInterfaceCopyAll,
    SCNetworkInterfaceGetBSDName,
    SCNetworkInterfaceGetHardwareAddressString,
    SCNetworkInterfaceGetInterfaceType,
    SCNetworkInterfaceGetSupportedProtocolTypes,
)


def get_interface_info(_fd: int, iface) -> dict[str, Any]:
    r = SCNetworkInterfaceCopyAll()
    if iface:
        for scnetworkinterface in r:
            if str(SCNetworkInterfaceGetBSDName(scnetworkinterface)) == iface:
                return do_get_interface_info(scnetworkinterface)
    elif len(r) == 1:
        return do_get_interface_info(r[0])
    return {}


def do_get_interface_info(scnetworkinterface) -> dict[str, Any]:
    info: dict[str, Any] = {}
    bsdname = SCNetworkInterfaceGetBSDName(scnetworkinterface)
    if bsdname:
        info["name"] = str(bsdname)
    hw = SCNetworkInterfaceGetHardwareAddressString(scnetworkinterface)
    if hw:
        info["hardware-address"] = str(hw)
    t = SCNetworkInterfaceGetInterfaceType(scnetworkinterface)
    if t:
        info["adapter-type"] = str(t)
    # SCNetworkInterfaceGetLocalizedDisplayName
    p = SCNetworkInterfaceGetSupportedProtocolTypes(scnetworkinterface)
    if p:
        info["protocols"] = tuple(str(x) for x in p)
    return info


def get_socket_tcp_info(sock) -> dict[str, Any]:
    from ctypes import c_int8, c_uint8, c_uint32, c_int32, c_int64
    MACOS_TCP_INFO_FIELDS = (
        ("state", c_uint8),
        ("options", c_uint8),
        ("snd_wscale", c_uint8),
        ("rcv_wscale", c_uint8),
        ("flags", c_uint32),
        ("rto", c_uint32),
        ("snd_mss", c_uint32),
        ("rcv_mss", c_uint32),
        ("rttcur", c_uint32),
        ("srtt", c_uint32),
        ("rttvar", c_uint32),
        ("rttbest", c_uint32),
        ("snd_ssthresh", c_uint32),
        ("snd_cwnd", c_uint32),
        ("rcv_space", c_uint32),
        ("snd_wnd", c_uint32),
        ("snd_nxt", c_uint32),
        ("rcv_nxt", c_uint32),
        ("last_outif", c_int32),
        ("snd_sbbytes", c_int32),
        ("padding", c_int32),
        # aligned 8:
        ("txpackets", c_int64),
        ("txbytes", c_int64),
        ("txretransmitbytes", c_int64),
        ("txunacked", c_int64),
        ("rxpackets", c_int64),
        ("rxbytes", c_int64),
        ("rxduplicatebytes", c_int64),
        ("rxoutoforderbytes", c_int64),
        ("snd_bw", c_int64),
        ("synrexmits", c_int8),
        ("unused1", c_int8),
        ("unused2", c_int8),
        ("cell_rxpackets", c_int64),
        ("cell_rxbytes", c_int64),
        ("cell_txpackets", c_int64),
        ("cell_txbytes", c_int64),
        ("wifi_rxpackets", c_int64),
        ("wifi_rxbytes", c_int64),
        ("wifi_txpackets", c_int64),
        ("wifi_txbytes", c_int64),
        ("wired_rxpackets", c_int64),
        ("wired_rxbytes", c_int64),
        ("wired_txpackets", c_int64),
        ("wired_txbytes", c_int64),
    )
    from xpra.net.socket_util import get_sockopt_tcp_info
    TCP_INFO = 0x200
    return get_sockopt_tcp_info(sock, TCP_INFO, MACOS_TCP_INFO_FIELDS)


def get_tcp_info(sock) -> dict[str, Any]:
    info = get_socket_tcp_info(sock)
    SO_NWRITE = 0x1024  # Get number of bytes currently in send socket buffer
    # actually gives the sum of (unsent data + sent-but-not-ACK-ed data)
    v = sock.getsockopt(socket.SOL_SOCKET, SO_NWRITE)
    info["sndbuf_unacked"] = v
    return info


def main():
    r = SCNetworkInterfaceCopyAll()
    print("%i interfaces:" % len(r))
    for scnetworkinterface in r:
        info = do_get_interface_info(scnetworkinterface)
        print(info)


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
from importlib import import_module

from xpra.os_util import is_admin
from xpra.platform import program_context
from xpra.util.config import get_user_config_file, get_system_conf_file
from xpra.scripts.config import InitExit
from xpra.exit_codes import ExitCode, ExitValue


def main(args) -> ExitValue:
    with program_context("Configure", "Configure"):
        if is_admin():
            conf = get_system_conf_file()
        else:
            conf = get_user_config_file()
        subcommand = args[0] if args else "home"
        if subcommand == "reset":
            import datetime
            now = datetime.datetime.now()
            with open(conf, "w", encoding="utf8") as f:
                f.write("# this file was reset on "+now.strftime("%Y-%m-%d %H:%M:%S"))
            return ExitCode.OK
        elif subcommand == "backup":
            if not os.path.exists(conf):
                print(f"# {conf!r} does not exist yet")
                return ExitCode.FILE_NOT_FOUND
            bak = conf[-5:]+".bak"
            with open(conf, "r", encoding="utf8") as read:
                with open(bak, "w", encoding="utf8") as write:
                    write.write(read.read())
            return ExitCode.OK
        elif subcommand == "show":
            if not os.path.exists(conf):
                print(f"# {conf!r} does not exist yet")
            else:
                with open(conf, "r", encoding="utf8") as f:
                    print(f.read())
            return ExitCode.OK
        else:
            if any(not str.isalnum(x) for x in subcommand):
                raise ValueError("invalid characters found in subcommand")
            try:
                mod = import_module(f"xpra.gtk.configure.{subcommand}")
            except ImportError:
                mod = None
            if not mod:
                raise InitExit(ExitCode.FILE_NOT_FOUND, f"unknown configure subcommand {subcommand!r}")
            return mod.main(args[1:])


if __name__ == "__main__":
    import sys
    sys.exit(int(main(sys.argv[1:])))



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
from typing import Any

from xpra.util.io import get_util_logger


def log_screen_sizes(root_w, root_h, sizes):
    try:
        do_log_screen_sizes(root_w, root_h, sizes)
    except Exception as e:
        get_util_logger().warn("failed to parse screen size information: %s", e, exc_info=True)


def prettify_plug_name(s, default="") -> str:
    if not s:
        return default
    try:
        s = s.decode("utf8")
    except (AttributeError, UnicodeDecodeError):
        pass
    # prettify strings on win32
    s = re.sub(r"[0-9\.]*\\", "-", s).lstrip("-")
    if s.startswith("WinSta-"):
        s = s[len("WinSta-"):]
    # ie: "(Standard monitor types) DELL ..."
    if s.startswith("(") and s.lower().find("standard") < s.find(") "):
        s = s.split(") ", 1)[1]
    if s == "0":
        s = default
    return s


def do_log_screen_sizes(root_w, root_h, sizes):
    from xpra.log import Logger
    log = Logger("screen")
    # old format, used by some clients (android):
    if not isinstance(sizes, (tuple, list)):
        return
    if any(True for x in sizes if not isinstance(x, (tuple, list))):
        return

    def dpi(size_pixels, size_mm):
        if size_mm == 0:
            return 0
        return round(size_pixels * 254 / size_mm / 10)

    def add_workarea(info, wx, wy, ww, wh):
        info.append("workarea: %4ix%-4i" % (ww, wh))
        if wx != 0 or wy != 0:
            # log position if not (0, 0)
            info.append("at %4ix%-4i" % (wx, wy))

    if len(sizes) != 1:
        log.warn("Warning: more than one screen found")
        log.warn(" this is not supported")
        log("do_log_screen_sizes(%i, %i, %s)", root_w, root_h, sizes)
        return
    s = sizes[0]
    if len(s) < 10:
        log.info(" %s", s)
        return
    # more detailed output:
    display_name, width, height, width_mm, height_mm, monitors, work_x, work_y, work_width, work_height = s[:10]
    # always log plug name:
    info = ["%s" % prettify_plug_name(display_name)]
    if width != root_w or height != root_h:
        # log plug dimensions if not the same as display (root):
        info.append("%ix%i" % (width, height))
    sdpix = dpi(width, width_mm)
    sdpiy = dpi(height, height_mm)
    info.append("(%ix%i mm - DPI: %ix%i)" % (width_mm, height_mm, sdpix, sdpiy))

    if work_width != width or work_height != height or work_x != 0 or work_y != 0:
        add_workarea(info, work_x, work_y, work_width, work_height)
    log.info("  " + " ".join(info))
    # sort monitors from left to right, top to bottom:
    monitors_distances = []
    for m in monitors:
        plug_x, plug_y = m[1:3]
        monitors_distances.append((plug_x + plug_y * width, m))
    sorted_monitors = [x[1] for x in sorted(monitors_distances)]
    for i, m in enumerate(sorted_monitors, start=1):
        if len(m) < 7:
            log.info("    %s", m)
            continue
        plug_name, plug_x, plug_y, plug_width, plug_height, plug_width_mm, plug_height_mm = m[:7]
        default_name = "monitor %i" % i
        info = ['%-16s' % prettify_plug_name(plug_name, default_name)]
        if plug_width != width or plug_height != height or plug_x != 0 or plug_y != 0:
            info.append("%4ix%-4i" % (plug_width, plug_height))
            if plug_x != 0 or plug_y != 0 or len(sorted_monitors) > 1:
                info.append("at %4ix%-4i" % (plug_x, plug_y))
        if (plug_width_mm != width_mm or plug_height_mm != height_mm) and (plug_width_mm > 0 or plug_height_mm > 0):
            dpix = dpi(plug_width, plug_width_mm)
            dpiy = dpi(plug_height, plug_height_mm)
            dpistr = ""
            if sdpix != dpix or sdpiy != dpiy or len(sorted_monitors) > 1:
                dpistr = " - DPI: %ix%i" % (dpix, dpiy)
            info.append("(%3ix%-3i mm%s)" % (plug_width_mm, plug_height_mm, dpistr))
        if len(m) >= 11:
            dwork_x, dwork_y, dwork_width, dwork_height = m[7:11]
            # only show it again if different from the screen workarea
            if dwork_x != work_x or dwork_y != work_y or dwork_width != work_width or dwork_height != work_height:
                add_workarea(info, dwork_x, dwork_y, dwork_width, dwork_height)
        if len(sorted_monitors) == 1 and len(info) == 1 and info[0].strip() in ("Canvas", "DUMMY0"):
            # no point in logging just `Canvas` on its own
            continue
        istr = (" ".join(info)).rstrip(" ")
        if len(monitors) == 1 and istr.lower() in ("unknown unknown", "0", "1", default_name, "screen", "monitor"):
            # a single monitor with no real name,
            # so don't bother showing it:
            continue
        log.info("    " + istr)


def get_screen_info(screen_sizes) -> dict[int, dict[str, Any]]:
    # same format as above
    if not screen_sizes:
        return {}
    info: dict[int, dict[str, Any]] = {}
    for i, x in enumerate(screen_sizes):
        if not isinstance(x, (tuple, list)):
            continue
        sinfo: dict[str, Any] = info.setdefault(i, {})
        sinfo["display"] = x[0]
        if len(x) >= 3:
            sinfo["size"] = x[1], x[2]
        if len(x) >= 5:
            sinfo["size_mm"] = x[3], x[4]
        if len(x) >= 6:
            monitors = x[5]
            for j, monitor in enumerate(monitors):
                if len(monitor) >= 7:
                    minfo: dict[str, Any] = sinfo.setdefault("monitor", {}).setdefault(j, {})
                    for k, v in {
                        "name": monitor[0],
                        "geometry": monitor[1:5],
                        "size_mm": monitor[5:7],
                    }.items():
                        minfo[k] = v
        if len(x) >= 10:
            sinfo["workarea"] = x[6:10]
    return info



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from collections.abc import Sequence

from xpra.util.env import envbool
from xpra.platform import platform_import

# defaults which may be overridden by platform_import:
CAN_DAEMONIZE: bool = True
REINIT_WINDOWS: bool = False
AUTOSTART: bool = False

INPUT_DEVICES: Sequence[str] = ("auto",)

SOURCE: Sequence[str] = ()

SYSTEM_PROXY_SOCKET: str = os.environ.get("XPRA_SYSTEM_PROXY_SOCKET", "/run/xpra/system")

CLIPBOARDS: Sequence[str] = ()
CLIPBOARD_WANT_TARGETS: bool = envbool("XPRA_CLIPBOARD_WANT_TARGETS")
CLIPBOARD_GREEDY: bool = envbool("XPRA_CLIPBOARD_GREEDY")
CLIPBOARD_PREFERRED_TARGETS: Sequence[str] = ("UTF8_STRING", "TEXT", "STRING", "text/plain")

EXECUTABLE_EXTENSION: str = ""

OPEN_COMMAND: Sequence[str] = ()

COMMAND_SIGNALS: Sequence[str] = ()

DEFAULT_START_ENV: Sequence[str] = ()
DEFAULT_ENV: Sequence[str] = ()

# DEFAULT_SSH_COMMAND = "paramiko"
DEFAULT_SSH_COMMAND: str = "ssh -x"
DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS: Sequence[tuple[str, str, str]] = (
    ("pactl", "set-default-sink", "Xpra-Speaker"),
    ("pactl", "set-default-source", "Xpra-Mic-Source"),
)

SOCKET_OPTIONS: Sequence[str] = (
    "SO_BROADCAST", "SO_RCVLOWAT",
    "SO_DONTROUTE", "SO_ERROR", "SO_EXCLUSIVEADDRUSE",
    "SO_KEEPALIVE", "SO_LINGER", "SO_OOBINLINE", "SO_RCVBUF",
    "SO_RCVTIMEO", "SO_REUSEADDR", "SO_REUSEPORT",
    "SO_SNDBUF", "SO_SNDTIMEO", "SO_TIMEOUT", "SO_TYPE",
)
IP_OPTIONS: Sequence[str] = (
    # "IP_MULTICAST_IF", "IP_MULTICAST_LOOP", "IP_MULTICAST_TTL",
    "IP_DONTFRAG", "IP_OPTIONS", "IP_RECVLCLIFADDR",
    "IP_RECVPKTINFO", "IP_TOS", "IP_TTL",
)
TCP_OPTIONS: Sequence[str] = ("TCP_NODELAY", "TCP_MAXSEG", "TCP_KEEPALIVE")

_features_list_: Sequence[str] = (
    "AUTOSTART",
    "CAN_DAEMONIZE",
    "REINIT_WINDOWS",
    "COMMAND_SIGNALS",
    "SOURCE",
    "DEFAULT_ENV",
    "DEFAULT_START_ENV",
    "DEFAULT_SSH_COMMAND",
    "DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS",
    "CLIPBOARDS",
    "CLIPBOARD_WANT_TARGETS",
    "CLIPBOARD_GREEDY",
    "CLIPBOARD_PREFERRED_TARGETS",
    "EXECUTABLE_EXTENSION",
    "INPUT_DEVICES",
    "SYSTEM_PROXY_SOCKET",
    "OPEN_COMMAND",
    "SOCKET_OPTIONS",
    "IP_OPTIONS",
    "TCP_OPTIONS",
)

platform_import(globals(), "features", False,
                *_features_list_)


def main():
    from xpra.util.str_fn import print_nested_dict
    from xpra.platform import program_context
    with program_context("Features-Info", "Features Info"):
        d = {}
        for k in _features_list_:
            d[k] = globals()[k]
        print_nested_dict(d)


if __name__ == "__main__":  # pragma: no cover
    main()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from xpra.os_util import OSX, POSIX, gi_import
from xpra.util.system import is_Wayland
from xpra.client.gtk3.client_base import GTKXpraClient
from xpra.client.gtk3.window import ClientWindow
from xpra.platform.gui import get_xdpi, get_ydpi

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GObject = gi_import("GObject")


class XpraClient(GTKXpraClient):
    ClientWindowClass = ClientWindow

    def __repr__(self):  # pylint: disable=arguments-differ
        return "gtk3.client"

    def client_type(self) -> str:
        return "Python/GTK3"

    def client_toolkit(self) -> str:
        if POSIX and not OSX:
            backend = os.environ.get("GDK_BACKEND", "")
            if not backend and is_Wayland():
                backend = "Wayland"
            if backend:
                # capitalize, ie: "x11" -> "X11"
                backend = backend[0].upper() + backend[1:]
                return f"GTK3 {backend}"
        return "GTK3"

    def get_notifier_classes(self) -> list[type]:
        ncs = super().get_notifier_classes()

        def nwarn(notifier_name: str, err: Exception) -> None:
            from xpra.log import Logger
            log = Logger("gtk", "client", "notify")
            log("get_notifier_classes()", exc_info=True)
            log.warn(f"Warning: failed to load the {notifier_name} notifier")
            log.warn(f" {err}")

        if not OSX:
            # pylint: disable=import-outside-toplevel
            try:
                from xpra.client.gtk3.notifier import GINotifier
                ncs.append(GINotifier)
            except Exception as e:
                nwarn("GObject", e)
        try:
            from xpra.gtk.notifier import GTKNotifier
            ncs.append(GTKNotifier)
        except Exception as e:
            nwarn("GTK", e)
        return ncs

    # noinspection PyMethodMayBeStatic
    def get_screen_resolution(self) -> int:
        screen = Gdk.Screen.get_default()
        if not screen:
            # wayland?
            return -1
        return round(screen.get_resolution())

    def get_xdpi(self) -> int:
        xdpi = get_xdpi()
        if xdpi > 0:
            return xdpi
        return round(self.get_screen_resolution())

    def get_ydpi(self) -> int:
        ydpi = get_ydpi()
        if ydpi > 0:
            return ydpi
        return round(self.get_screen_resolution())

    def get_mouse_position(self) -> tuple[int, int]:
        # with GTK3, we can get None values!
        root = self.get_root_window()
        if not root:
            return -1, -1
        p = root.get_pointer()[-3:-1]
        return self.cp(p[0] or 0, p[1] or 0)


GObject.type_register(XpraClient)



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from ctypes import WinDLL, WINFUNCTYPE, oledll, c_int, c_char, byref, cast
from ctypes.wintypes import BOOL, LPVOID, LPCWSTR, LPCVOID, LPOLESTR

dsound = WinDLL("dsound", use_last_error=True)
DirectSoundEnumerateW = dsound.DirectSoundEnumerateW
DirectSoundCaptureEnumerate = dsound.DirectSoundCaptureEnumerateW
GetDeviceID = dsound.GetDeviceID
# DEFINE_GUID(DSDEVID_DefaultPlayback,     0xDEF00000,0x9C6D,0x47Ed,0xAA,0xF1,0x4D,0xDA,0x8F,0x2B,0x5C,0x03);
# DEFINE_GUID(DSDEVID_DefaultCapture,      0xDEF00001,0x9C6D,0x47Ed,0xAA,0xF1,0x4D,0xDA,0x8F,0x2B,0x5C,0x03);
# DEFINE_GUID(DSDEVID_DefaultVoicePlayback,0xDEF00002,0x9C6D,0x47Ed,0xAA,0xF1,0x4D,0xDA,0x8F,0x2B,0x5C,0x03);
# DEFINE_GUID(DSDEVID_DefaultVoiceCapture, 0xDEF00003,0x9C6D,0x47ED,0xAA,0xF1,0x4D,0xDA,0x8F,0x2B,0x5C,0x03);

LPDSENUMCALLBACK = WINFUNCTYPE(BOOL, LPVOID, LPCWSTR, LPCWSTR, LPCVOID)
StringFromGUID2 = oledll.ole32.StringFromGUID2
StringFromGUID2.restype = c_int
StringFromGUID2.argtypes = [LPVOID, LPOLESTR, c_int]


def _enum_devices(fn):
    devices = []

    def cb_enum(lpGUID, lpszDesc, _lpszDrvName, _):
        dev = ""
        if lpGUID is not None:
            # noinspection PyTypeChecker
            buftype = c_char * 256
            # noinspection PyCallingNonCallable
            buf = buftype()
            pbuf = byref(buf)
            if StringFromGUID2(lpGUID, cast(pbuf, LPOLESTR), 256):
                dev = buf.value
        devices.append((dev, lpszDesc))
        return True

    fn(LPDSENUMCALLBACK(cb_enum), None)
    return devices


def get_devices():
    return _enum_devices(DirectSoundEnumerateW)


def get_capture_devices():
    return _enum_devices(DirectSoundCaptureEnumerate)


def main():
    import sys
    from xpra.platform import program_context
    from xpra.log import Logger, enable_color, consume_verbose_argv
    with program_context("Audio Device Info", "Audio Device Info"):
        enable_color()
        consume_verbose_argv(sys.argv, "audio")
        log = Logger("win32", "audio")
        log.info("")
        log.info("Capture Devices:")
        for k, v in get_capture_devices():
            log.info("* %-40s : %s", v, k)
        log.info("")
        log.info("All Devices:")
        for guid, name in get_devices():
            log.info("* %-40s : %s", name, guid)


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import glob

from xpra.os_util import POSIX, getuid
from xpra.util.env import osexpand, envbool
from xpra.util.io import load_binary_file

CLEAN_SESSION_FILES = envbool("XPRA_CLEAN_SESSION_FILES", True)


def get_session_dir(mode: str, sessions_dir: str, display_name: str, uid: int) -> str:
    session_dir = osexpand(os.path.join(sessions_dir, display_name.lstrip(":")), uid=uid)
    if not os.path.exists(session_dir):
        ROOT = POSIX and getuid() == 0
        ROOT_FALLBACK = ("/run/xpra", "/var/run/xpra", "/tmp")
        if ROOT and uid == 0 and not any(session_dir.startswith(x) for x in ROOT_FALLBACK):
            # there is usually no $XDG_RUNTIME_DIR when running as root
            # and even if there was, that's probably not a good path to use,
            # so try to find a more suitable directory we can use:
            for d in ROOT_FALLBACK:
                if os.path.exists(d):
                    if mode == "proxy" and (display_name or "").lstrip(":").split(",")[0] == "14500":
                        # stash the system-wide proxy session files in a 'proxy' subdirectory:
                        return os.path.join(d, "proxy")
                    # otherwise just use the display as subdirectory name:
                    return os.path.join(d, (display_name or "").lstrip(":"))
    return session_dir


def make_session_dir(mode: str, sessions_dir: str, display_name: str, uid: int = 0, gid: int = 0) -> str:
    session_dir = get_session_dir(mode, sessions_dir, display_name, uid)
    if not os.path.exists(session_dir):
        try:
            os.makedirs(session_dir, 0o750, exist_ok=True)
        except OSError:
            import tempfile
            session_dir = osexpand(os.path.join(tempfile.gettempdir(), display_name.lstrip(":")))
            os.makedirs(session_dir, 0o750, exist_ok=True)
        ROOT = POSIX and getuid() == 0
        mismatch = ROOT and uid != 0 or gid != 0
        if mismatch and (session_dir.startswith("/run/user/") or session_dir.startswith("/run/xpra/")):
            os.lchown(session_dir, uid, gid)
    return session_dir


def session_file_path(filename: str) -> str:
    session_dir = os.environ.get("XPRA_SESSION_DIR")
    if session_dir is None:
        raise RuntimeError("'XPRA_SESSION_DIR' must be set to use this function")
    return os.path.join(session_dir, filename)


def load_session_file(filename: str) -> bytes:
    return load_binary_file(session_file_path(filename))


def save_session_file(filename: str, contents: str | bytes, uid: int = -1, gid: int = -1) -> str:
    if not os.environ.get("XPRA_SESSION_DIR"):
        return ""
    if not isinstance(contents, bytes):
        contents = str(contents).encode("utf8")
    assert contents
    path = session_file_path(filename)
    try:
        with open(path, "wb+") as f:
            if POSIX:
                os.fchmod(f.fileno(), 0o640)
                if getuid() == 0 and uid >= 0 and gid >= 0:
                    os.fchown(f.fileno(), uid, gid)
            f.write(contents)
    except OSError as e:
        from xpra.log import Logger
        log = Logger("server")
        log("save_session_file", exc_info=True)
        log.error(f"Error saving session file {path!r}")
        log.estr(e)
    return path


def rm_session_dir(warn: bool = True) -> None:
    session_dir = os.environ.get("XPRA_SESSION_DIR")
    if not session_dir or not os.path.exists(session_dir):
        return
    from xpra.log import Logger
    log = Logger("server")
    try:
        session_files = os.listdir(session_dir)
    except OSError as e:
        log("os.listdir(%s)", session_dir, exc_info=True)
        if warn:
            log.error(f"Error: cannot access {session_dir!r}")
            log.estr(e)
        return
    if session_files:
        if warn:
            log.info(f"session directory {session_dir!r} was not removed")
            log.info(" because it still contains some files:")
            for f in session_files:
                extra = " (directory)" if os.path.isdir(os.path.join(session_dir, f)) else ""
                log.info(f" {f!r}{extra}")
        return
    try:
        os.rmdir(session_dir)
    except OSError as e:
        log = Logger("server")
        log(f"rmdir({session_dir})", exc_info=True)
        log.error(f"Error: failed to remove session directory {session_dir!r}")
        log.estr(e)


def clean_session_files(*filenames) -> None:
    if not CLEAN_SESSION_FILES:
        return
    for filename in filenames:
        path = session_file_path(filename)
        if filename.find("*") >= 0 or filename.find("?") >= 0:
            for p in glob.glob(path):
                clean_session_path(p)
        else:
            clean_session_path(path)
    rm_session_dir(False)


def clean_session_path(path) -> None:
    from xpra.log import Logger
    log = Logger("server")
    log(f"clean_session_path({path})")
    if os.path.exists(path):
        try:
            if os.path.isdir(path):
                os.rmdir(path)
            else:
                os.unlink(path)
        except OSError as e:
            log(f"clean_session_path({path})", exc_info=True)
            log.error(f"Error removing session path {path}")
            log.estr(e)
            if os.path.isdir(path):
                files = os.listdir(path)
                if files:
                    log.error(" this directory still contains some files:")
                    for file in files:
                        finfo = repr(file)
                        try:
                            if os.path.islink(file):
                                finfo += " -> "+repr(os.readlink(file))
                        except OSError:
                            pass
                        log.error(f" {finfo}")



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import struct
import unittest
import binascii

from xpra.util.env import OSEnvContext
from unit.test_util import LoggerSilencer


class XSettingsTest(unittest.TestCase):

    def test_xsettings(self):
        from xpra.x11.xsettings_prop import (
            xsettings_to_bytes, bytes_to_xsettings,
            get_local_byteorder,
            XSettingsType,
            )
        for v in (None, 10, "", (), [], (1, 2, 3), [1, ]):
            try:
                xsettings_to_bytes(v)
            except (ValueError, TypeError):
                continue
            else:
                raise RuntimeError(f"should not be able to parse {v!r}")

        for v in (None, 10, struct.pack(b"=BBBBII", 0, 0, 0, 0, 100, 2)):
            #setting_type, _, name_len = struct.unpack(b"=BBH", d[pos:pos + 4])
            try:
                assert not bytes_to_xsettings(v)
            except (ValueError, TypeError):
                continue
            else:
                raise RuntimeError(f"should not be able to parse {v!r}")

        for DEBUG_XSETTINGS in (True, False):
            with OSEnvContext():
                os.environ["XPRA_XSETTINGS_DEBUG"] = str(int(DEBUG_XSETTINGS))
                serial = 1
                data = b""
                l = len(data)
                v = struct.pack(b"=BBBBII", get_local_byteorder(), 0, 0, 0, serial, l)+data+b"\0"
                v1 = bytes_to_xsettings(v)
                assert v
                # get from cache:
                v2 = bytes_to_xsettings(v)
                assert v1==v2

                # test all types, set then get:
                # setting_type, prop_name, value, last_change_serial = setting
                settings = (
                    (XSettingsType.Integer, "int1", 1, 0),
                    (XSettingsType.String, "str1", "1", 0),
                    (XSettingsType.Color, "color1", (128, 128, 64, 32), 0),
                    )
                serial = 2
                data = xsettings_to_bytes((serial, settings))
                assert data
                # parse it back:
                v = bytes_to_xsettings(data)
                rserial, rsettings = v
                assert rserial==serial
                assert len(rsettings)==len(settings)
        from xpra.x11 import xsettings_prop
        with LoggerSilencer(xsettings_prop):
            # test error handling:
            for settings in (
                (
                    # invalid color causes exception
                    (XSettingsType.Color, "bad-color", (128, ), 0),
                ),
                (
                    # invalid setting type is skipped with an error message:
                    (255, "invalid-setting-type", 0, 0),
                ),
            ):
                serial = 3
                data = xsettings_to_bytes((serial, settings))
                assert data
                v = bytes_to_xsettings(data)
                rserial, rsettings = v
                assert rserial==serial
                assert len(rsettings)==0
            # parsing an invalid data type (9) should fail:
            hexdata = b"000000000200000001000000090004007374723100000000010000003100000000"
            data = binascii.unhexlify(hexdata)
            v = bytes_to_xsettings(data)
            rserial, rsettings = v
            assert len(rsettings)==0


def main():
    #can only work with an X11 server
    from xpra.os_util import OSX, POSIX
    if POSIX and not OSX:
        unittest.main()

if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# This class simply hooks the current GDK display into
# the core X11 bindings.

from xpra.util.system import is_X11
from xpra.os_util import gi_import
from xpra.scripts.config import InitException
from xpra.x11.bindings.xlib cimport Display
from xpra.x11.bindings.display_source cimport set_display   # pylint: disable=syntax-error
from xpra.x11.bindings.display_source import set_display_name  # @UnresolvedImport


###################################
# Headers, python magic
###################################
cdef extern from "gtk-3.0/gdk/gdk.h":
    pass

cdef extern from "gtk-3.0/gdk/gdktypes.h":
    ctypedef struct GdkDisplay:
        pass

cdef extern from "gtk-3.0/gdk/gdkdisplay.h":
    GdkDisplay *gdk_display_get_default()

cdef extern from "gtk-3.0/gdk/gdkx.h":
    Display *gdk_x11_display_get_xdisplay(GdkDisplay *display)
    const char *gdk_display_get_name(GdkDisplay *display)


#this import magic will make the window.get_xid() available!
GdkX11 = gi_import("GdkX11")
if GdkX11 is None:
    raise RuntimeError("could not load GdkX11 3.0")

display = None
def init_gdk_display_source() -> None:
    global display
    import os
    backend = os.environ.get("GDK_BACKEND", "")
    if backend!="x11" and not is_X11():
        raise InitException("cannot use X11 bindings with %s and GTK3 (buggy)" % (backend or "non-X11",))
    if display:
        return
    Gdk = gi_import("Gdk")
    cdef GdkDisplay* gdk_display = gdk_display_get_default()
    if not gdk_display:
        raise InitException("cannot access display '%s'" % os.environ.get("DISPLAY", ""))
    #this next line actually ensures Gdk is initialized, somehow
    root = Gdk.get_default_root_window()
    if root is None:
        raise RuntimeError("could not get the default root window")
    display = Gdk.Display.get_default()
    #now we can get a display:
    cdef Display *x11_display = gdk_x11_display_get_xdisplay(gdk_display)
    set_display(x11_display)
    name = gdk_display_get_name(gdk_display).decode()
    set_display_name(name)

def close_gdk_display_source() -> None:
    #this triggers the garbage collection of the Display object:
    global display
    display = None



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import shutil
import sys
import time
import tempfile
import unittest
import subprocess
from typing import NoReturn

from xpra.util.str_fn import repr_ellipsized, bytestostr
from xpra.util.env import envint, envbool, osexpand, OSEnvContext
from xpra.os_util import (
    OSX, POSIX,
)
from xpra.util.io import pollwait
from xpra.platform.paths import get_xpra_command
from xpra.platform.dotxpra import DISPLAY_PREFIX
from xpra.scripts.main import X11_SOCKET_DIR, stat_display_socket
from xpra.scripts.config import get_defaults

from xpra.log import Logger

log = Logger("test")

XPRA_TEST_DEBUG = envbool("XPRA_TEST_DEBUG", False)
XVFB_TIMEOUT = envint("XPRA_TEST_XVFB_TIMEOUT", 8)
DELETE_TEMP_FILES = envbool("XPRA_DELETE_TEMP_FILES", True)
SHOW_XORG_OUTPUT = envbool("XPRA_SHOW_XORG_OUTPUT", False)
TEST_XVFB_COMMAND = os.environ.get("XPRA_TEST_VFB_COMMAND", "Xvfb")


def show_proc_pipes(proc) -> None:
    def showfile(fileobj, filetype="stdout") -> None:
        if fileobj and fileobj.name and os.path.exists(fileobj.name):
            log.warn("contents of %s file '%s':", filetype, fileobj.name)
            try:
                with fileobj as f:
                    f.seek(0)
                    for line in f:
                        log.warn(" %s", line.decode("latin1").rstrip("\n\r"))
            except Exception as e:
                log.error("Error: failed to read '%s': %s", fileobj.name, e)
        else:
            log.warn("no %s file", filetype)

    showfile(proc.stdout_file, "stdout")
    showfile(proc.stderr_file, "stderr")


def show_proc_error(proc, msg) -> NoReturn:
    if not proc:
        raise Exception("command failed to start: %s" % msg)
    log.warn("%s failed:", proc.command)
    log.warn("returncode=%s", proc.poll())
    show_proc_pipes(proc)
    raise Exception(msg + " command=%s" % proc.command)


class DisplayContext(OSEnvContext):

    def __init__(self):
        OSEnvContext.__init__(self)
        self.xvfb_process = None

    def __enter__(self):
        OSEnvContext.__enter__(self)
        if POSIX and not OSX:
            ProcessTestUtil.setUpClass()
            self.stu = ProcessTestUtil()
            self.stu.setUp()
            self.xvfb_process = self.stu.start_Xvfb()
            os.environ["GDK_BACKEND"] = "x11"
            os.environ["DISPLAY"] = self.xvfb_process.display or ""
            from xpra.x11.gtk.display_source import init_gdk_display_source
            init_gdk_display_source()

    def __exit__(self, *_args):
        if self.xvfb_process:
            self.xvfb_process.terminate()
            self.xvfb_process = None
            self.stu.tearDown()
            ProcessTestUtil.tearDownClass()
        OSEnvContext.__exit__(self)

    def __repr__(self):
        return "DisplayContext"


class ProcessTestUtil(unittest.TestCase):

    def __init__(self, methodName='runTest'):
        if not hasattr(self, "runTest"):
            self.runTest = None
        unittest.TestCase.__init__(self, methodName)

    @classmethod
    def setUpClass(cls):
        cls.display_start = 100
        cls.temp_files = []
        cls.processes = []
        cls.xauthority_temp = None  #tempfile.NamedTemporaryFile(prefix="xpra-test.", suffix=".xauth", delete=False)
        #cls.xauthority_temp.close()
        #os.environ["XAUTHORITY"] = os.path.expanduser(cls.xauthority_temp.name)
        cls.default_env = os.environ.copy()
        cls.default_env.update({
            "XPRA_LOG_DIR": tempfile.gettempdir(),
            "XPRA_NOTTY": "1",
            "XPRA_WAIT_FOR_INPUT": "0",
        })
        cls.default_config = get_defaults()
        log("setUpClass(%s) default_env=%s", cls, cls.default_env)

    @classmethod
    def tearDownClass(cls):
        if cls.xauthority_temp:
            os.unlink(cls.xauthority_temp.name)
        cls.stop_commands()
        if DELETE_TEMP_FILES:
            for x in cls.temp_files:
                try:
                    os.unlink(x)
                except OSError:
                    pass

    @classmethod
    def stop_commands(cls):
        for x in cls.processes:
            try:
                if x.poll() is None:
                    x.terminate()
                stdout_file = getattr(x, "stdout_file", None)
                if stdout_file:
                    try:
                        stdout_file.close()
                    except OSError:
                        pass
                stderr_file = getattr(x, "stderr_file", None)
                if stderr_file:
                    try:
                        stderr_file.close()
                    except OSError:
                        pass
            except OSError:
                log.error("failed to stop subprocess %s", x)

        def get_wait_for():
            return tuple(proc for proc in cls.processes if proc.poll() is None)

        wait_for = get_wait_for()
        start = time.monotonic()
        while wait_for and time.monotonic() - start < 5:
            if len(wait_for) == 1:
                pollwait(wait_for[0])
            else:
                time.sleep(1)
            wait_for = get_wait_for()

    def setUp(self):
        os.environ.clear()
        os.environ.update(self.default_env)

    @classmethod
    def get_default_run_env(cls):
        env = dict((k, v) for k, v in cls.default_env.items() if k.startswith("XPRA") or k in (
            "HOME", "HOSTNAME", "SHELL", "TERM",
            "USER", "USERNAME", "PATH",
            "XAUTHORITY", "PWD",
            "PYTHONPATH", "SYSTEMROOT",
            "DBUS_SESSION_BUS_ADDRESS",
            "XDG_RUNTIME_DIR",
        ))
        log("get_default_run_env() env(%s)=%s", repr_ellipsized(cls.default_env), env)
        env["NO_AT_BRIDGE"] = "1"
        return env

    def get_run_env(self):
        return self.get_default_run_env()

    @classmethod
    def which(cls, cmd):
        return shutil.which(cmd) or cmd

    def run_command(self, command, **kwargs):
        return self.class_run_command(command, **kwargs)

    @classmethod
    def class_run_command(cls, command, **kwargs):
        if "env" not in kwargs:
            kwargs["env"] = cls.get_default_run_env()
        stdout_file = stderr_file = None
        if isinstance(command, (list, tuple)):
            strcommand = " ".join("'%s'" % x for x in command)
        else:
            strcommand = command
        if XPRA_TEST_DEBUG:
            log("************************")
            log("class_run_command(%s, %s)", command, repr_ellipsized(str(kwargs), 80))
            log("************************")
        else:
            if "stdout" not in kwargs:
                stdout_file = cls._temp_file(prefix="xpra-stdout-")
                kwargs["stdout"] = stdout_file
                log("stdout: %s for %s", stdout_file.name, strcommand)
            if "stderr" not in kwargs:
                stderr_file = cls._temp_file(prefix="xpra-stderr-")
                kwargs["stderr"] = stderr_file
                log("stderr: %s for %s", stderr_file.name, strcommand)
        try:
            log("class_run_command%s", (command, kwargs))
            proc = subprocess.Popen(args=command, **kwargs)
            proc.command = command
            proc.stdout_file = stdout_file
            proc.stderr_file = stderr_file
        except OSError as e:
            log.warn("class_run_command(%s, %s) %s", command, kwargs, e)
            raise
        cls.processes.append(proc)
        return proc

    @classmethod
    def get_xpra_cmd(cls):
        cmd = get_xpra_command()
        if cmd == ["xpra"]:
            cmd = [bytestostr(cls.which("xpra"))]
        pyexename = os.environ.get("PYTHON", "python3")
        exe = bytestostr(cmd[0])
        if exe.endswith(".exe"):
            exe = exe[:-4]
        if not (exe.endswith("python") or exe.endswith(pyexename) or exe == "coverage"):
            #prepend python / python3:
            cmd = [pyexename] + cmd
        return cmd

    @classmethod
    def show_proc_pipes(cls, proc):
        show_proc_pipes(proc)

    @classmethod
    def show_proc_error(cls, proc, msg):
        show_proc_error(proc, msg)

    @classmethod
    def get_command_output(cls, command, env=None, **kwargs):
        proc = cls.class_run_command(command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        out = proc.communicate()[0]
        return out

    @classmethod
    def _temp_file(cls, data=None, prefix="xpra-"):
        f = tempfile.NamedTemporaryFile(prefix=prefix, delete=False)
        if data:
            f.file.write(data)
        f.file.flush()
        cls.temp_files.append(f.name)
        return f

    def delete_temp_file(self, f):
        try:
            self.temp_files.remove(f.name)
        except ValueError:
            pass
        try:
            os.unlink(f.name)
        except OSError as e:
            log.error("Error deleting temp file '%s': %s", f.name, e)

    @classmethod
    def find_X11_display_numbers(cls):
        #use X11 sockets:
        X11_displays = set()
        if POSIX and os.path.exists(X11_SOCKET_DIR):
            for x in os.listdir(X11_SOCKET_DIR):
                if not x.startswith("X"):
                    continue
                try:
                    dno = int(x[1:])
                except ValueError:
                    continue
                sock_path = os.path.join(X11_SOCKET_DIR, x)
                if stat_display_socket(sock_path):
                    X11_displays.add(dno)
        return X11_displays

    @classmethod
    def find_X11_displays(cls):
        return ["%s%i" % (DISPLAY_PREFIX, x) for x in cls.find_X11_display_numbers()]

    @classmethod
    def find_free_display_no(cls, exclude=()):
        #X11 sockets:
        X11_displays = cls.find_X11_displays()
        start = cls.display_start % 10000
        for i in range(start, 20000):
            display = "%s%s" % (DISPLAY_PREFIX, i)
            if display in exclude:
                continue
            if display in X11_displays:
                continue
            cls.display_start += 100
            return i
        raise Exception("failed to find any free displays!")

    @classmethod
    def find_free_display(cls):
        return "%s%i" % (DISPLAY_PREFIX, cls.find_free_display_no())

    @classmethod
    def start_Xvfb(cls, display=None, screens=((1024, 768),), depth=24, extensions=("Composite",)) -> subprocess.Popen:
        assert POSIX
        if display is None:
            display = cls.find_free_display()
        env = {}
        for x in list(os.environ.keys()):
            if x in (
                    "LOGNAME", "USER", "PATH", "LANG", "TERM",
                    "HOME", "USERNAME", "PYTHONPATH", "HOSTNAME",
                    #"XAUTHORITY",
            ):  #DBUS_SESSION_BUS_ADDRESS
                #keep it
                env[x] = os.environ.get(x)
        real_display = os.environ.get("DISPLAY")
        cmd = [cls.which(TEST_XVFB_COMMAND)]
        is_xephyr = TEST_XVFB_COMMAND.find("Xephyr") >= 0
        if is_xephyr:
            if len(screens) > 1 or not real_display:
                #we can't use Xephyr for multi-screen
                cmd = [cls.which("Xvfb")]
            elif real_display:
                env["DISPLAY"] = real_display
        for ext in extensions:
            if ext.startswith("-"):
                cmd += ["-extension", ext[1:]]
            else:
                cmd += ["+extension", ext]
        cmd += ["-nolisten", "tcp", "-noreset"]
        #"-auth", self.default_env["XAUTHORITY"]]
        depth_str = "%i+%i" % (depth, 32)
        for i, screen in enumerate(screens):
            (w, h) = screen
            cmd += ["-screen"]
            if not is_xephyr:
                cmd += ["%i" % i]
            cmd += ["%ix%ix%s" % (w, h, depth_str)]
        cmd.append(display)
        env["XPRA_LOG_DIR"] = tempfile.gettempdir()
        cmd_expanded = [osexpand(v) for v in cmd]
        cmdstr = " ".join("'%s'" % x for x in cmd_expanded)
        if SHOW_XORG_OUTPUT:
            stdout = sys.stdout
            stderr = sys.stderr
        else:
            stdout = cls._temp_file(prefix="Xorg-stdout-")
            log("stdout=%s for %s", stdout.name, cmd)
            stderr = cls._temp_file(prefix="Xorg-stderr-")
            log("stderr=%s for %s", stderr.name, cmd)
        xvfb = cls.class_run_command(cmd_expanded, env=env, stdout=stdout, stderr=stderr)
        xvfb.display = display
        time.sleep(1)
        log("xvfb(%s)=%s" % (cmdstr, xvfb))
        if pollwait(xvfb, XVFB_TIMEOUT) is not None:
            show_proc_error(xvfb, "xvfb command \"%s\" failed and returned %s" % (cmdstr, xvfb.poll()))
        return xvfb



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.log import Logger, consume_verbose_argv

log = Logger("network", "util", "win32")


def get_interface_info(_fd, iface):
    from xpra.platform.win32.comtypes_util import CIMV2_Query
    query = "SELECT * FROM Win32_NetworkAdapter WHERE GUID='%s'" % iface
    try:
        with CIMV2_Query(query) as res:
            log("ExecQuery(%s) returned %i rows", query, res.Count)
            if res.Count == 1:
                for r in res:
                    props = {}
                    for k, ik, conv in (
                            ("AdapterType", "adapter-type", str),
                            ("Caption", "caption", str),
                            ("Description", "description", str),
                            ("DeviceID", "id", int),
                            ("GUID", "GUID", str),
                            ("Index", "index", int),
                            ("Name", "name", str),
                            ("ProductName", "product-name", str),
                            ("Speed", "speed", int),
                    ):
                        try:
                            v = conv(r.Properties_[k].Value)
                        except Exception as e:
                            log.error("Error retrieving '%s' from network adapter record:", k)
                            log.estr(e)
                        else:
                            props[ik] = v
                    log("get_interface_info(%s)=%s" % (iface, props))
                    return props
    except Exception as e:
        log("get_interface_info(%s)", iface, exc_info=True)
        from xpra.util.env import first_time
        if first_time("win32-network-query"):
            log.info("cannot query network interface:")
            log.info(" %s", e)
    return {}


def get_tcp_info(_sock):  # pylint: disable=unused-argument
    """
    #not implemented yet!
    #the functions below would require administrator privileges:
    from ctypes import WinDLL, POINTER, c_void_p, c_ubyte
    from ctypes.wintypes import ULONG, UINT
    PUCHAR = POINTER(c_ubyte)
    iphlpapi = WinDLL("Iphlpapi", use_last_error=True)
    TCP_ESTATS_TYPE = UINT
    PMIB_TCPROW = c_void_p
    SetPerTcpConnectionEStats = iphlpapi.SetPerTcpConnectionEStats
    SetPerTcpConnectionEStats.restype = ULONG
    SetPerTcpConnectionEStats.argtypes = (PMIB_TCPROW, TCP_ESTATS_TYPE,
                                          PUCHAR, ULONG, ULONG, ULONG)
    GetPerTcpConnectionEStats = iphlpapi.GetPerTcpConnectionEStats
    GetPerTcpConnectionEStats.restype = ULONG
    GetPerTcpConnectionEStats.argtypes = (PMIB_TCPROW, TCP_ESTATS_TYPE,
                                          PUCHAR, ULONG, ULONG,
                                          PUCHAR, ULONG, ULONG,
                                          PUCHAR, ULONG, ULONG)
    """
    return {}


def main():
    import sys
    from xpra.platform import program_context
    with program_context("Network-Speed", "Network Speed Query Tool"):
        consume_verbose_argv(sys.argv, "win32")
        from xpra.net.net_util import get_interfaces
        from xpra.util.stats import std_unit
        interfaces = get_interfaces()
        for iface in interfaces:
            speed = get_interface_info(0, iface).get("speed", 0)
            try:
                v = int(speed)
                s = "%sbps" % std_unit(v)
                print("%s : %s" % (iface, s))
            except ValueError:
                log.error("Error: parsing speed value '%s'", speed, exc_info=True)


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


def get_server_base_classes() -> tuple[type, ...]:
    from xpra.server import features
    from xpra.server.core import ServerCore
    classes: list[type] = [ServerCore]
    if features.dbus:
        from xpra.server.mixins.dbus import DbusServer
        classes.append(DbusServer)
    if features.control:
        from xpra.server.mixins.controlcommands import ServerBaseControlCommands
        classes.append(ServerBaseControlCommands)
    if features.notifications:
        from xpra.server.mixins.notification import NotificationForwarder
        classes.append(NotificationForwarder)
    if features.webcam:
        from xpra.server.mixins.webcam import WebcamServer
        classes.append(WebcamServer)
    if features.clipboard:
        from xpra.server.mixins.clipboard import ClipboardServer
        classes.append(ClipboardServer)
    if features.audio:
        from xpra.server.mixins.audio import AudioServer
        classes.append(AudioServer)
    if features.fileprint:
        from xpra.server.mixins.fileprint import FilePrintServer
        classes.append(FilePrintServer)
    if features.mmap:
        from xpra.server.mixins.mmap import MMAP_Server
        classes.append(MMAP_Server)
    if features.input_devices:
        from xpra.server.mixins.input import InputServer
        classes.append(InputServer)
    if features.encoding:
        from xpra.server.mixins.encoding import EncodingServer
        classes.append(EncodingServer)
    if features.logging:
        from xpra.server.mixins.logging import LoggingServer
        classes.append(LoggingServer)
    if features.network_state:
        from xpra.server.mixins.networkstate import NetworkStateServer
        classes.append(NetworkStateServer)
    if features.ssh:
        from xpra.server.mixins.ssh_agent import SshAgent
        classes.append(SshAgent)
    if features.http:
        from xpra.server.mixins.http import HttpServer
        classes.append(HttpServer)
    if features.shell:
        from xpra.server.mixins.shell import ShellServer
        classes.append(ShellServer)
    if features.display:
        from xpra.server.mixins.display import DisplayManager
        classes.append(DisplayManager)
    if features.cursors:
        from xpra.server.mixins.cursors import CursorManager
        classes.append(CursorManager)
    if features.windows:
        from xpra.server.mixins.window import WindowServer
        classes.append(WindowServer)
    if features.commands:
        from xpra.server.mixins.child_command import ChildCommandServer
        classes.append(ChildCommandServer)
    return tuple(classes)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import time
import unittest

from xpra.util.objects import typedict
from xpra.os_util import OSX, POSIX
from xpra.util.io import pollwait, which
from unit.server_test_util import ServerTestUtil, log


class X11ServerTest(ServerTestUtil):

	def start_test_xvfb_and_server(self, *args):
		display = self.find_free_display()
		xvfb = self.start_Xvfb(display)
		time.sleep(1)
		assert display in self.find_X11_displays()
		#start server using this display:
		server = self.check_fast_start_server(display, "--use-display=yes", *args)
		return display, xvfb, server

	def start_test_server(self, *args):
		display = self.find_free_display()
		log("starting test server on %s", display)
		server = self.check_fast_start_server(display, *args)
		assert display in self.find_X11_displays()
		return display, server


	def test_display_reuse(self):
		display, server = self.start_test_server()
		#make sure we cannot start another server on the same display:
		saved_spe = self.show_proc_error
		def raise_exception_no_log(*args):
			raise Exception("%s" % args)
		#suspend process error logging:
		self.show_proc_error = raise_exception_no_log
		try:
			try:
				log("should not be able to start another test server on %s", display)
				self.check_fast_start_server(display)
			except Exception:
				pass
			else:
				raise Exception("server using the same display should have failed to start")
		finally:
			self.show_proc_error = saved_spe
		assert server.poll() is None, "server should not have terminated"
		#tell the server to exit and leave the display behind:
		log("asking the server to exit")
		self.check_stop_server(server, "exit", display)
		del server
		assert display not in self.dotxpra.displays(), "server socket for display should have been removed"
		#now we can start it again using "--use-display"
		log("start a new server on the same display")
		server = self.check_fast_start_server(display, "--use-display=yes")
		assert display in self.dotxpra.displays(), "server display not found"
		#shut it down now
		self.check_stop_server(server, "stop", display)
		#the socket may be removed before the vfb is stopped,
		#so wait a little:
		for _ in range(20):
			if display not in self.find_X11_displays():
				return
			time.sleep(1)
		raise RuntimeError(f"the display {display!r} should have been killed by the server shutdown")


	def test_existing_Xvfb(self):
		display, xvfb, server = self.start_test_xvfb_and_server()
		self.check_stop_server(server, "stop", display)
		time.sleep(1)
		assert pollwait(xvfb, 2) is None, "the Xvfb should not have been killed by xpra shutting down!"
		xvfb.terminate()


	def test_dbus(self):
		dbus_send = which("dbus-send")
		if not dbus_send:
			print("Warning: dbus test skipped, 'dbus-send' not found")
			return
		#make sure this is recorded,
		#so kill the xvfb *after* we exit
		display, xvfb, server = self.start_test_xvfb_and_server("--start=xterm", "--start=xterm", "-d", "dbus")	#2 windows
		info = self.get_server_info(display)
		assert info
		tinfo = typedict(info)
		dstr = display.lstrip(":")
		env = self.get_run_env()
		env["DISPLAY"] = display
		#use the dbus environment from the server:
		for k,v in tinfo.items():
			if k.startswith("env.DBUS_"):
				env[k[4:]] = v
		#unchecked calls:
		for args in (
			("org.xpra.Server.Focus", "int32:1"),
			#("org.xpra.Server.Suspend", ),		#hangs?
			("org.xpra.Server.Resume", ),
			("org.xpra.Server.Ungrab", ),
			("org.xpra.Server.Start", "string:xterm"),
			("org.xpra.Server.StartChild", "string:xterm"),
			("org.xpra.Server.EnableDebug", "string:keyboard"),
			("org.xpra.Server.KeyPress", "int32:36"),		#Return key with default layout?
			("org.xpra.Server.KeyRelease", "int32:36"),		#Return key
			("org.xpra.Server.ClearKeysPressed", ),
			("org.xpra.Server.SetKeyboardRepeat", "int32:200", "int32:50"),
			("org.xpra.Server.DisableDebug", "string:keyboard"),
			("org.xpra.Server.ListClients", ),
			("org.xpra.Server.DetachClient", "string:nomatch"),
			("org.xpra.Server.DetachAllClients", ),
			("org.xpra.Server.GetAllInfo", ),
			("org.xpra.Server.GetInfo", "string:keyboard"),
			("org.xpra.Server.ListWindows", ),
			("org.xpra.Server.LockBatchDelay", "int32:1", "int32:10"),
			("org.xpra.Server.UnlockBatchDelay", "int32:1"),
			("org.xpra.Server.MovePointer", "int32:1", "int32:100", "int32:100"),
			("org.xpra.Server.MouseClick", "int32:1", "boolean:true"),
			("org.xpra.Server.MouseClick", "int32:1", "boolean:false"),
			("org.xpra.Server.MoveWindowToWorkspace", "int32:1", "int32:2"),
			("org.xpra.Server.RefreshAllWindows", ),
			("org.xpra.Server.RefreshWindow", "int32:1"),
			("org.xpra.Server.RefreshWindows", "array:int32:1:2:3"),
			("org.xpra.Server.ResetVideoRegion", "int32:1"),
			("org.xpra.Server.ResetWindowFilters", ),
			("org.xpra.Server.ResetXSettings", ),
			("org.xpra.Server.SendNotification", "int32:1", "string:title", "string:message", "string:*"),
			("org.xpra.Server.CloseNotification", "int32:1", "string:*"),
			("org.xpra.Server.SendUIClientCommand", "array:string:test"),
			("org.xpra.Server.SetClipboardProperties", "string:direction", "int32:1000", "int32:1000"),
			("org.xpra.Server.SetIdleTimeout", "int32:100"),
			("org.xpra.Server.SetLock", "string:true"),
			("org.xpra.Server.SetLock", "string:false"),
			("org.xpra.Server.SetLock", "string:auto"),
			("org.xpra.Server.SetSharing", "string:no"),
			("org.xpra.Server.SetScreenSize", "int32:1024", "int32:768"),
			("org.xpra.Server.SetUIDriver", "string:invalid"),
			("org.xpra.Server.SetVideoRegion", "int32:1", "int32:10", "int32:10", "int32:100", "int32:100"),
			("org.xpra.Server.SetVideoRegionDetection", "int32:1", "boolean:true"),
			("org.xpra.Server.SetVideoRegionDetection", "int32:1", "boolean:false"),
			("org.xpra.Server.SetVideoRegionEnabled", "int32:1", "boolean:true"),
			("org.xpra.Server.SetVideoRegionEnabled", "int32:1", "boolean:false"),
			#("org.xpra.Server.SetVideoRegionExclusionZones", ...
			("org.xpra.Server.SetWindowEncoding", "int32:1", "string:png"),
			("org.xpra.Server.SetWindowScaling", "int32:1", "string:200%"),
			("org.xpra.Server.SetWindowScalingControl", "int32:1", "string:auto"),
			("org.xpra.Server.SetWindowScalingControl", "int32:1", "string:50"),
			("org.xpra.Server.SetWorkarea", "int32:10", "int32:10", "int32:1000", "int32:740"),
			("org.xpra.Server.ShowAllWindows", ),
			("org.xpra.Server.SyncXvfb", ),
			("org.xpra.Server.SetDPI", "int32:144", "int32:144"),
			("org.xpra.Server.ToggleFeature", "string:bell", "string:off"),
			#generic dbus server queries:
			("org.xpra.Server.Get", "string:session_name"),
			("org.xpra.Server.GetAll", ),
			("org.xpra.Server.Set", "string:session_name", "string:foo"),
			):
			cmd = [dbus_send, "--session", "--type=method_call",
					"--dest=org.xpra.Server%s" % dstr, "/org/xpra/Server"] + list(args)
			proc = self.run_command(cmd, env=env)
			assert pollwait(proc, 20) is not None, "dbus-send is taking too long: %s" % (cmd,)
		#properties using interface org.freedesktop.DBus.Properties:
		for args in (
			("Get", "string:idle-timeout"),
			("Get", "string:does-not-exist"),
			("GetAll", ),
			("Set", "string:idle-timeout", "int32:20"),
			("Set", "string:does-not-exist", "string:irrelevant"),
			):
			cmd = [dbus_send, "--session", "--type=method_call", "--print-reply",
					"--dest=org.xpra.Server%s" % dstr, "/org/xpra/Server",
					"org.freedesktop.DBus.Properties.%s" % args[0], "string:org.xpra.Server%s" % dstr
					] + list(args[1:])
			proc = self.run_command(cmd, env=env)
			assert pollwait(proc, 20) is not None, "dbus-send is taking too long: %s" % (cmd,)

		self.check_stop_server(server, "exit", display)
		xvfb.terminate()


def main():
	if POSIX and not OSX:
		unittest.main()


if __name__ == '__main__':
	main()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
import os.path
from typing import Any, TypeAlias
from collections.abc import Callable, Iterable, Sequence

from xpra.util.str_fn import Ellipsizer
from xpra.util.env import envbool
from xpra.os_util import getuid, OSX, POSIX
from xpra.util.io import get_proc_cmdline
from xpra.platform.paths import get_user_conf_dirs, get_system_conf_dirs
from xpra.log import Logger

log = Logger("window", "util")

GUESS_CONTENT = envbool("XPRA_GUESS_CONTENT", True)
DEFAULT_CONTENT_TYPE = os.environ.get("XPRA_DEFAULT_CONTENT_TYPE", "")
CONTENT_TYPE_DEFS = os.environ.get("XPRA_CONTENT_TYPE_DEFS", "")

ConfDict: TypeAlias = dict[str, Any]
ConfParser: TypeAlias = Callable[[Sequence[str]], ConfDict]


def getprop(window, prop: str):
    with log.trap_error("Error querying %r on %r", prop, window):
        if prop not in window.get_property_names():
            log("no '%s' property on window %s", prop, window)
            return None
        return window.get_property(prop)


################################################################
# generic file parsing functions
################################################################


def _load_dict_file(filename: str, parser: ConfParser) -> ConfDict:
    # filter out comments and remove line endings
    lines = []
    with open(filename, encoding="utf8") as f:
        for line in f:
            line = line.rstrip("\n\r")
            if line.startswith("#") or not line.strip():
                continue
            lines.append(line)
    log("_load_dict_file(%s)=%s", filename, Ellipsizer(lines))
    return parser(lines)


def _load_dict_dir(d: str, parser: ConfParser) -> ConfDict:
    # load all the .conf files from the directory
    if not os.path.exists(d) or not os.path.isdir(d):
        log("load_content_categories_dir(%s) directory not found", d)
        return {}
    v = {}
    for f in sorted(os.listdir(d)):
        if f.endswith(".conf"):
            cc_file = os.path.join(d, f)
            if os.path.isfile(cc_file):
                try:
                    v.update(_load_dict_file(cc_file, parser))
                except Exception as e:
                    log("_load_dict_dir(%s)", cc_file, exc_info=True)
                    log.error("Error loading file data from '%s'", cc_file)
                    log.estr(e)
    log("_load_dict_dir(%s)=%s", d, v)
    return v


def _load_dict_dirs(dirname: str, parser: ConfParser) -> ConfDict:
    if not GUESS_CONTENT:
        return {}
    # finds all the ".conf" files from the dirname specified
    # and calls `load` on them.
    # looks for system and user conf dirs
    values = {}
    for d in get_system_conf_dirs():
        v = _load_dict_dir(os.path.join(d, dirname), parser)
        values.update(v)
    if not POSIX or getuid() > 0:
        for d in get_user_conf_dirs():
            v = _load_dict_dir(os.path.join(d, dirname), parser)
            values.update(v)
    return values


################################################################
# `content-type` mapping:
################################################################

content_type_defs: ConfDict | None = None


def load_content_type_defs() -> ConfDict:
    global content_type_defs
    if content_type_defs is None:
        content_type_defs = _load_dict_dirs("content-type", parse_content_types)
        if GUESS_CONTENT:
            # add env defs:
            content_type_defs.update(parse_content_types(CONTENT_TYPE_DEFS.split(",")))
    return content_type_defs


def parse_content_types(lines) -> dict[str, dict[Any, tuple[str, str]]]:
    defs: dict[str, dict[Any, tuple[str, str]]] = {}
    for line in lines:
        if not line:
            continue
        parts = line.rsplit("=", 1)
        # ie: "title:helloworld=text   #some comments here" -> "title:helloworld", "text   #some comments here"
        if len(parts) != 2:
            log.warn("Warning: invalid content-type definition")
            log.warn(f" {line!r} is missing a '='")
            continue
        match_str, content_type = parts
        parts = match_str.split(":", 1)
        # ie: "title:helloworld" -> "title", "helloworld"
        if len(parts) != 2:
            log.warn("Warning: invalid content-type definition")
            log.warn(f" match string {match_str!r} is missing a ':'")
            continue
        # ignore comments:
        # "text    #some comments here" > "text"
        content_type = content_type.split(":")[0].strip()
        prop_name, regex = parts
        try:
            c = re.compile(regex)
        except Exception as e:
            log.warn("Warning: invalid regular expression")
            log.warn(f" match string {regex!r}:")
            log.warn(f" {e}")
            continue
        else:
            defs.setdefault(prop_name, {})[c] = (regex, content_type)
            log("%16s matching '%s' is %s", prop_name, regex, content_type)
    return defs


def get_content_type_properties() -> Iterable[str]:
    """ returns the list of window properties which can be used
        to guess the content-type.
    """
    return load_content_type_defs().keys()


def guess_content_type_from_defs(window) -> str:
    load_content_type_defs()
    assert content_type_defs is not None
    for prop_name, defs in content_type_defs.items():
        if prop_name not in window.get_property_names():
            continue
        prop_value = window.get_property(prop_name)
        # some properties return lists of values,
        # in which case we try to match any of them:
        log("guess_content_type_from_defs(%s) prop(%s)=%s", window, prop_name, prop_value)
        if isinstance(prop_value, (list, tuple)):
            values = prop_value
        else:
            values = [prop_value]
        for value in values:
            for regex, match_data in defs.items():
                if regex.search(str(value)):
                    regex_str, content_type = match_data
                    log("guess_content_type(%s) found match: property=%s, regex=%s, content-type=%s",
                        window, prop_name, regex_str, content_type)
                    return content_type
    return ""


################################################################
# `content-categories` mapping:
################################################################


def parse_content_categories_file(lines) -> ConfDict:
    d: dict[str, str] = {}
    for line in lines:
        parts = line.rsplit(":", 1)
        # ie: "title:helloworld=text   #some comments here" -> "title:helloworld", "text   #some comments here"
        if len(parts) != 2:
            log.warn("Warning: invalid content-type definition")
            log.warn(" %r is missing a '='", line)
            continue
        category, content_type = parts
        d[category.strip("\t ").lower()] = content_type.strip("\t ")
    log("parse_content_categories_file(%s)=%s", lines, d)
    return d


def load_categories_to_type() -> ConfDict:
    return _load_dict_dirs("content-categories", parse_content_categories_file)


################################################################
# command mapping: using menu data
################################################################

command_to_type: dict[str, str] | None = None


def load_command_to_type() -> dict[str, str]:
    if not GUESS_CONTENT:
        return {}
    global command_to_type
    if command_to_type is not None:
        return command_to_type
    command_to_type = {}
    from xpra.server.menu_provider import get_menu_provider
    xdg_menu = get_menu_provider().get_menu_data(remove_icons=True)
    categories_to_type = load_categories_to_type()
    log("load_command_to_type() xdg_menu=%s, categories_to_type=%s", xdg_menu, categories_to_type)
    if xdg_menu and categories_to_type:
        for category, category_props in xdg_menu.items():
            log("category %s: %s", category, Ellipsizer(category_props))
            entries = category_props.get("Entries", {})
            for name, props in entries.items():
                command = str(props.get("TryExec") or props.get("Exec") or "")
                categories = props.get("Categories")
                log("Entry '%s': command=%s, categories=%s", name, command, categories)
                if command and categories:
                    for c in categories:
                        ctype = categories_to_type.get(c.lower())
                        if not ctype:
                            # try a more fuzzy match:
                            for category_name, ct in categories_to_type.items():
                                if c.lower().find(category_name) >= 0:
                                    ctype = ct
                                    break
                        if ctype:
                            cmd = os.path.basename(command.split(" ")[0])
                            if cmd:
                                command_to_type[cmd] = str(ctype)
                                break
    log("load_command_to_type()=%s", command_to_type)
    return command_to_type


def guess_content_type_from_command(window) -> str:
    if POSIX and not OSX:
        command = getprop(window, "command") or ""
        log(f"guess_content_type_from_command({window}) command={command}")
        if command:
            ctt = load_command_to_type()
            commands = [command]
            for x in command.split("\0"):
                if x and x not in commands:
                    commands.append(x)
            for x in command.split(" "):
                if x and x not in commands:
                    commands.append(x)
            for x in commands:
                cmd = os.path.basename(x)
                ctype = ctt.get(cmd)
                log("content-type(%s)=%s", cmd, ctype)
                if ctype:
                    return ctype
    return ""


################################################################
# `content-parent` mapping:
################################################################

def parse_content_parent(lines) -> dict[str, str]:
    v = {}
    for line in lines:
        parts = line.split(":", 1)
        if len(parts) == 2:
            v[parts[0].strip()] = parts[1].strip()
    return v


parent_to_type = None


def get_parent_to_type() -> dict[str, Any]:
    global parent_to_type
    if parent_to_type is None:
        parent_to_type = _load_dict_dirs("content-parent", parse_content_parent)
    return parent_to_type


def guess_content_type_from_parent(window) -> str:
    ppid = getprop(window, "ppid")
    if not ppid:
        return ""
    return guess_content_from_parent_pid(ppid)


def guess_content_from_parent_pid(ppid: int) -> str:
    parent_command = get_proc_cmdline(ppid)
    if not parent_command:
        return ""
    executable = os.path.basename(parent_command[0])
    pt = get_parent_to_type()
    return pt.get(executable, "")


def guess_content_type(window) -> str:
    if not GUESS_CONTENT:
        return DEFAULT_CONTENT_TYPE
    return guess_content_type_from_defs(window) or \
        guess_content_type_from_command(window) or \
        guess_content_type_from_parent(window) or \
        DEFAULT_CONTENT_TYPE


def main():
    # pylint: disable=import-outside-toplevel
    import sys
    args = sys.argv[1:]
    while "-v" in args:
        args.remove("-v")
        log.enable_debug()
    if not args:
        print("usage: %s pids" % sys.argv[0])
        return
    for ppid in args:
        c = guess_content_from_parent_pid(int(ppid))
        print(f"guess_content_from_parent_pid({ppid})={c}")


if __name__ == "__main__":  # pragma: no cover
    main()
